/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <termios.h>
#include <unistd.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "types.h"
#include "errorRoutines.h"
#include "misc.h"
#include "translate.h"
#include "ix86_state.h"
#include "memory.h"
#include "syscall.h"
#include "operands.h"
#include "helper.h"

#include "func.h"

/***************************************************************************
 ***************************************************************************/
#ifdef MICRO_OPS
static long long occurances[OP_MAX];

int              micro_ops_this_cycle;

#define MAX_OPS_DIST 1024
static long long opsDist[MAX_OPS_DIST+2];

#define OPS_DIST_PROBE(NOPS)		\
  (assert(NOPS >= 0),                   \
   ((NOPS) > MAX_OPS_DIST)		\
   ? opsDist[MAX_OPS_DIST+1]++		\
   : opsDist[NOPS]++)
#endif

/***************************************************************************
 ***************************************************************************/
/* track number of refs, useful for gauging input sizes */
static unsigned long maxInsts  = 0;
static char          *baseName = NULL;

/* state for the functional simulation of the processor core */
static cpu_t          nState;


/***************************************************************************
 ***************************************************************************/
static
void
DumpDist(long long *dist, int len, char *fname)
{
	int       i;
	long long sum,
	          subSum,
	          subSum1;
	FILE      *distFile;
	char      buf[128];

	if (baseName) {
		strcpy(buf, baseName);
		strcat(buf, fname);
	} else
		strcpy(buf, fname);

	if (!(distFile = fopen(buf, "w")))
		fatal("could not open distribution result file `%s'", buf);
  
	subSum = 0; subSum1 = 0; sum = 0;
	for (i = 0; i <= len+1; i++)
		sum += dist[i];
  
	fprintf(distFile, "value\t\toccurances\n");
	for (i = 0; i < len; i++)
		fprintf(distFile, "%6d  %15.0f   # %f%%  %f%%  %f%%\n",
			i,
			(double)dist[i],
			(double)dist[i]/(double)sum,
			(double)(subSum += dist[i])/(double)sum,
			(double)(subSum1 += (i==0 ? 0 : dist[i]))/(double)(sum-dist[0]));
	fprintf(distFile, "%6d  %15.0f   # %f%%  %f%%  %d+ actually\n",
		len, (double)dist[len], (double)dist[len]/(double)sum,
		(double)(subSum += dist[len])/(double)sum, len);
	fprintf(distFile, "%6d  %15.0f   # %f%%  %f%%  <0 actually\n",
		len+1, (double)dist[len+1], (double)dist[len+1]/(double)sum,
		(double)(subSum += dist[len+1])/(double)sum);
	fclose(distFile);

}/*** end DumpDist() ***/



/***************************************************************************
 ***************************************************************************/
void
SimOptions(int argc, char **argv)
{
	char c;

	/* parse options */
	NewOpt();
	while ((c = GetOpt(argc, argv, ALL_OPTIONS)) != EOF) {
		switch (c) {
		case 'm':
		{
			if (sscanf(optArg, "%lu",&maxInsts) != 1)
				fatal("could not parse maximum instructions parameter");
			info("simulating a maximum of %lu instructions.",maxInsts);
		}
			break;
		case 'W':
			baseName = strdup(optArg);
			break;
		}
	}

}/*** end SimOptions() ***/




/***************************************************************************
 ***************************************************************************/
void
SimConfig(void)
{
	/* register an error handler */
	RegisterErrorFn(DumpState);

}/*** end SimConfig() ***/




/***************************************************************************
 ***************************************************************************/
void
SimStats(FILE *stream)
{

#ifdef MICRO_OPS
	int i;
	FILE *ops_file;
	char buf[128];
	
	if (baseName) {
		strcpy(buf, baseName);
		strcat(buf, ".ops_file");
	}
	else
		strcpy(buf, "oops_file");

	if ((ops_file = fopen(buf,"w")) == NULL)
		info("couldn't open ops_file '%s'",buf);

	for (i = 0; i < OP_MAX; i++)
		fprintf(ops_file,"%20s: %f\n",op2name[i],(double)occurances[i]);

	DumpDist(opsDist, MAX_OPS_DIST, ".ops-dist");
#endif

}/*** end SimStats() ***/





/***************************************************************************
 ***************************************************************************/
static
void
halt(void)
{
	fatal("encountered halt instruction");

}/*** end halt() ***/



/***************************************************************************
 ***************************************************************************/
static
void
dummy(iInst_t inst)
{
	fatal("unimplemented opcode:0x%x -- %s\n",inst.opcodeByte,op2name[inst.op]);

}/*** end dummy() ***/



/***************************************************************************
 ***************************************************************************/
void
SimMain(int argc, char **argv, char **envp)
{
	iInst_t inst;
	enum Opcode op;

#define TEMP_VARS_ONLY
#include "ix86.def"
#undef TEMP_VARS_ONLY

#ifdef MICRO_OPS
	int i;
#endif
	Bit32u   nextPC;

	if (!baseName)
		fatal("set base name (-Wbase_name) before running functional simulation");

	info("starting functional simulation");
	numInsts = 0;

	/* initialize the decode tables */
	InitDecoder();        /* in translate.c */

	/*
	   dumpOpcodeInfo(stdout);
	   */

	/* initialize the processor state */
	initState(&nState);          /* in ix86_state.c */
	InitPreciseState(&nState);   /* in state.c */

	/* load the program text and data, set up environment */
	LoadProgram(argc, argv, envp, &nState);

	nState.eip = progEntry - 8;
	nextPC = progEntry;

#ifdef MICRO_OPS
	for (i = 0; i< OP_MAX; i++)
		occurances[i] = 0;
#endif

        while (1) {
		
		if (dumpStats) {
			DumpState(stderr,&nState);
			DumpFinalStats(stderr);
			dumpStats = FALSE;
		}
		  
		nState.eip = nextPC;

		/*
		 * figure out what the next instruction is and
		 * execute it in the case statement below
		 */
		inst = getInst(nState.eip);

		/* default next pc is just the sequentially next instruction */
		nextPC = nState.eip + inst.iSize;

		if (verbose)
			DisassembleInst(inst,stderr, nState.eip);
		else {
			if ((verbInst && (numInsts == verbInst)) || (verbAddr && (nState.eip == verbAddr))) {
				info("turning of verbostity");
				DisassembleInst(inst,stderr, nState.eip);
				verbose = TRUE;
			}
		}
		  
		op = inst.op;

#include "non-spec.defines"

#ifdef MICRO_OPS
		micro_ops_this_cycle = 1;
#endif

		switch (op) {
			/* load it in */

#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE)	\
		          case OP: EXPR;  \
	                           break;
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef DEFGROUP
                    default:
	                 error("bogus opcode `%d' encountered", op);
		}

		/* verify that all flag values are booleans */
		assert((CF == 0) || (CF == 1));
		assert((PF == 0) || (PF == 1));
		assert((AF == 0) || (AF == 1));
		assert((ZF == 0) || (ZF == 1));
		assert((SF == 0) || (SF == 1));
		assert((OF == 0) || (OF == 1));

#include "non-spec.undefines"



		/* keep an instruction count */
		numInsts++;

#ifdef MICRO_OPS
		occurances[op]++; 
		OPS_DIST_PROBE(micro_ops_this_cycle);
#endif

		if (maxInsts && (numInsts >= maxInsts))
			longjmp(exitBuf, /* exitcode */0);
	}

}/*** end SimMain() ***/
