/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <string.h>

#include "func.h"
#include "misc.h"
#include "memory.h"
#include "ix86_state.h"
#include "errorRoutines.h"

char           *x86RegName[NUM_REGS];

/* enum Opcode -> description string */
char *op2name[OP_MAX] = {
	"XXXXXX", /* NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) NAME,
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFFU
#undef DEFINST
#undef DEFLINK
};


/***********************************************************************************
 ***********************************************************************************/
void
DumpState(FILE *stream, cpu_t *state)
{
	int i;

	assert(stream);

	fprintf(stream, "Processor state:\n");
	fprintf(stream, "    PC: 0x%08lx  Insts: %.0f  Cycles: %.0f\n",
		state->eip, (double)nnumInsts, (double)numCycles);
	fprintf(stream, "   brk: 0x%08lx   minSP: 0x%08lx\n",
		breakPoint, minSP);
	fprintf(stream, "    lastSyscall: %ld\n", lastSyscall);

	if(state)
	  {
		  for (i=0; i<NUM_REGS; i += 2) {
			  fprintf(stream, "    %s: %12ld/0x%08lx",
				  x86RegName[i],   *(state->gen_reg32[i]),   *(state->gen_reg32[i]));
			  fprintf(stream, "    %s: %12ld/0x%08lx\n",
				  x86RegName[i+1], *(state->gen_reg32[i+1]), *(state->gen_reg32[i+1]));
		  }
	  }
	else
		fprintf(stream,"STATE to DumpState is NULL.\n");

}


#if(0)
/***********************************************************************************
 ***********************************************************************************/
void
DumpErrorState(FILE *stream,cpu_t *state)
{
	DumpState(stream,state);
}
#endif

static int first = 1;

/***********************************************************************************
 ***********************************************************************************/
void
initState(cpu_t *state)
{
	int i;

	state->gen_reg8[BIT_REG_AL] = (Bit8u *) (((Bit8u *) &(state->eax)) + REG8L_OFFSET);
	state->gen_reg8[BIT_REG_CL] = (Bit8u *) (((Bit8u *) &(state->ecx)) + REG8L_OFFSET);
	state->gen_reg8[BIT_REG_DL] = (Bit8u *) (((Bit8u *) &(state->edx)) + REG8L_OFFSET);
	state->gen_reg8[BIT_REG_BL] = (Bit8u *) (((Bit8u *) &(state->ebx)) + REG8L_OFFSET);
	state->gen_reg8[BIT_REG_AH] = (Bit8u *) (((Bit8u *) &(state->eax)) + REG8H_OFFSET);
	state->gen_reg8[BIT_REG_CH] = (Bit8u *) (((Bit8u *) &(state->ecx)) + REG8H_OFFSET);
	state->gen_reg8[BIT_REG_DH] = (Bit8u *) (((Bit8u *) &(state->edx)) + REG8H_OFFSET);
	state->gen_reg8[BIT_REG_BH] = (Bit8u *) (((Bit8u *) &(state->ebx)) + REG8H_OFFSET);

	state->gen_reg16[BIT_REG_AX] = (Bit16u *) (((Bit8u *) &(state->eax)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_CX] = (Bit16u *) (((Bit8u *) &(state->ecx)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_DX] = (Bit16u *) (((Bit8u *) &(state->edx)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_BX] = (Bit16u *) (((Bit8u *) &(state->ebx)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_SP] = (Bit16u *) (((Bit8u *) &(state->esp)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_BP] = (Bit16u *) (((Bit8u *) &(state->ebp)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_SI] = (Bit16u *) (((Bit8u *) &(state->esi)) + REG16_OFFSET);
	state->gen_reg16[BIT_REG_DI] = (Bit16u *) (((Bit8u *) &(state->edi)) + REG16_OFFSET);

	state->gen_reg32[BIT_REG_EAX] = (Bit32u *) (&(state->eax));
	state->gen_reg32[BIT_REG_ECX] = (Bit32u *) (&(state->ecx));
	state->gen_reg32[BIT_REG_EDX] = (Bit32u *) (&(state->edx));
	state->gen_reg32[BIT_REG_EBX] = (Bit32u *) (&(state->ebx));
	state->gen_reg32[BIT_REG_ESP] = (Bit32u *) (&(state->esp));
	state->gen_reg32[BIT_REG_EBP] = (Bit32u *) (&(state->ebp));
	state->gen_reg32[BIT_REG_ESI] = (Bit32u *) (&(state->esi));
	state->gen_reg32[BIT_REG_EDI] = (Bit32u *) (&(state->edi));

	/* initialize naming structures if this is the first time we have been called */
	if(first)
	  {
		  first=0;

		  x86RegName[BIT_REG_EAX] = malloc(1+strlen("EAX")*sizeof(char));
		  x86RegName[BIT_REG_ECX] = malloc(1+strlen("ECX")*sizeof(char));
		  x86RegName[BIT_REG_EDX] = malloc(1+strlen("EDX")*sizeof(char));
		  x86RegName[BIT_REG_EBX] = malloc(1+strlen("EBX")*sizeof(char));
		  x86RegName[BIT_REG_ESP] = malloc(1+strlen("ESP")*sizeof(char));
		  x86RegName[BIT_REG_EBP] = malloc(1+strlen("EBP")*sizeof(char));
		  x86RegName[BIT_REG_ESI] = malloc(1+strlen("ESI")*sizeof(char));
		  x86RegName[BIT_REG_EDI] = malloc(1+strlen("EDI")*sizeof(char));

		  strcpy(x86RegName[BIT_REG_EAX],"EAX");
		  strcpy(x86RegName[BIT_REG_ECX],"ECX");
		  strcpy(x86RegName[BIT_REG_EDX],"EDX");
		  strcpy(x86RegName[BIT_REG_EBX],"EBX");
		  strcpy(x86RegName[BIT_REG_ESP],"ESP");
		  strcpy(x86RegName[BIT_REG_EBP],"EBP");
		  strcpy(x86RegName[BIT_REG_ESI],"ESI");
		  strcpy(x86RegName[BIT_REG_EDI],"EDI");
	  }

	/* initialize floating point unit */
	for (i = 0; i<8; i++)
	  {
		  state->fReg[i] = 0.0;
		  state->fTag[i] = FPU_TAG_EMPTY;
	  }
	state->CONTROL_REG = 0x037f;
	state->STATUS_REG  = 0x0000;

	/* verify that initial control register settings match Intel's specs */
	assert(state->CONTROL_IM == 1);
	assert(state->CONTROL_DM == 1);
	assert(state->CONTROL_ZM == 1);
	assert(state->CONTROL_OM == 1);
	assert(state->CONTROL_UM == 1);
	assert(state->CONTROL_PM == 1);
	assert(state->CONTROL_PC == 3);
	assert(state->CONTROL_RC == 0);
	assert(state->CONTROL_REG = 0x037f);
}


/***********************************************************************************
 ***********************************************************************************/
void
InitPreciseState(cpu_t *state)
{
	unsigned long i;

	for (i=0; i<NUM_REGS; i++)
	  {
		  /* don't mess with the stack pointer as it has already been initialized */
		  if (i != BIT_REG_ESP)
			  *(state->gen_reg32[i]) = 0;
	  }
}


/***********************************************************************************
 ***********************************************************************************/
void
DumpFinalStats(FILE *stream)
{
	time_t elapsedTime;

	elapsedTime = MAX(time((time_t *)NULL) - startTime, 1);
	info("simulation time: %s", ElapsedTime(elapsedTime));
	info("executed %.0f insts, %f insts/sec",
	     (double)numInsts, (double)numInsts/elapsedTime);
	info("  of which %.2f%% hit in the simulators decoded instruction cache",
	     (double)100*(double)((double)numDecodeCacheHits/(double)numInsts));
	info("executed %.0f cycles", (double)numCycles);
	info("average issue rate %f insts/cycle", (double)numInsts/numCycles);
	
	SimStats(stream);

	/* dump memory usage statistics */
	MemStats(stream);
}
