/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef IX86_STATE_H
#define IX86_STATE_H

#include "types.h"

enum sizes { BYTE, WORD, DOUBLE };
enum repeats { NO_REP, REPE, REPNE };
enum overrides { NO_OVER, ES_OVER, SS_OVER, CS_OVER, DS_OVER, FS_OVER, GS_OVER };

typedef struct {
	Boolean cf;
	Boolean pf;
	Boolean af;
	Boolean zf;
	Boolean sf;
	Boolean tf;
	Boolean if_;
	Boolean df;
	Boolean of;
	Bit8u   iopl;
	Boolean nt;
	Boolean rf;
	Boolean vm;
	Boolean ac;
} flags_reg_t;

typedef struct { /* selector_t */
	Bit16u value;   /* the 16bit value of the selector */
	/* the following fields are extracted from the value field in virtual
	   mode only.  They're used for sake of efficiency */
	Bit16u index;   /* 13bit index extracted from value in virtual mode */
	Bit8u  ti;      /* table indicator bit extracted from value */
	Bit8u  rpl;     /* RPL extracted from value */
} selector_t;

  /* the x86 maintains a descriptor cache for each segment register.  Upon
     loading a segment register, fields from the descriptor it selects are
     cached in the corresponding descriptor cache.  the descriptor cache
     structure is not available by applications, it is maintained soley
     by the CPU
   */
typedef struct { /* segment_attributes_t */
	Bit8u   dwordcount; /* parameter count for call gates only */
	Boolean d;         /* default size, 0=16bit ops & addrs, 1=32bit */
	Boolean p;         /* present bit */
	Bit8u   dpl;       /* descriptor privilege level */
	Boolean dtype;     /* descriptor type, 0=system, 1=application */
	Bit8u   type;      /* segment type, 3bit
			      0 = read-only data segment
			      1 = read-write data segment
			      2 = read-only expand down data segment
			      3 = read-write expand down data segment
			      4 = execute-only code segment
			      5 = execute-readable code segment
			      6 = execute-only conforming code segment
			      7 = execute-readable conforming code segment
			      */
	Boolean g;         /* limit granularity. 0=byte, 1=4K */
} segment_attr_t;

typedef struct {
	selector_t         selector;
	Bit32u             base;      /* base address, 32bit */
	Bit32u             limit;     /* limit, 16bit or 20bit */
	segment_attr_t     attributes; /* shadow register attributes */
} segment_reg_t;

typedef struct {
	Boolean pg; /* paging */
	Boolean cd; /* cache disable */
	Boolean nw; /* no write-through */
	Boolean am; /* alignment mask */
	Boolean wp; /* write-protect */
	Boolean ne; /* numerics exception */
	Boolean et; /* extension type */
	Boolean ts; /* task switched */
	Boolean em; /* emulate math coprocessor */
	Boolean mp; /* monitor coprocessor */
	Boolean pe; /* protected mode enable */
} cr0_t;

typedef Float64 FReg_t;

typedef int FloatTagWord_t;
#define FPU_TAG_VALID	0
#define FPU_TAG_ZERO	1
#define FPU_TAG_SPECIAL	2
#define FPU_TAG_EMPTY	3

typedef union {
	struct bits {
		unsigned int unused:16,
	        reserved2:3,
	        infinity_control:1,
	        rc:2,
	        pc:2,
                reserved1:2,
                pm:1,
                um:1,
                om:1,
                zm:1,
                dm:1,
                im:1;
	} bits_control;
	int reg;
} FloatControlReg_t;
#define ROUND_NEAR_EVEN 0x0
#define ROUND_DOWN      0x1
#define ROUND_UP        0x2
#define ROUND_CHOP      0x3
#define CONTROL_REG fControl.reg
#define CONTROL_IM fControl.bits_control.im
#define CONTROL_DM fControl.bits_control.dm
#define CONTROL_ZM fControl.bits_control.zm
#define CONTROL_OM fControl.bits_control.om
#define CONTROL_UM fControl.bits_control.um
#define CONTROL_PM fControl.bits_control.pm
#define CONTROL_PC fControl.bits_control.pc
#define CONTROL_RC fControl.bits_control.rc
#define CONTROL_REG fControl.reg

typedef union {
	struct {
		unsigned int unused:16,
	                     b:1,
	                     c3:1,
                             top:3,
                             c2:1,
                             c1:1,
                             c0:1,
                             es:1,
                             sf:1,
                             pe:1,
                             ue:1,
                             oe:1,
                             ze:1,
                             de:1,
                             ie:1;
	} bits_status;
	int reg;
} FloatStatusReg_t;
#define STATUS_IE  fStatus.bits_status.ie
#define STATUS_DE  fStatus.bits_status.de
#define STATUS_ZE  fStatus.bits_status.ze
#define STATUS_OE  fStatus.bits_status.oe
#define STATUS_UE  fStatus.bits_status.ue
#define STATUS_PE  fStatus.bits_status.pe
#define STATUS_SF  fStatus.bits_status.sf
#define STATUS_ES  fStatus.bits_status.es
#define STATUS_C3  fStatus.bits_status.c3
#define STATUS_C2  fStatus.bits_status.c2
#define STATUS_C1  fStatus.bits_status.c1
#define STATUS_C0  fStatus.bits_status.c0
#define STATUS_CC  ((fStatus.bits_status.c3 << 3) | (fStatus.bits_status.c2 << 2) | (fStatus.bits_status.c1 << 1)  | (fStatus.bits_status.c0 << 0)) 
#define STATUS_TOP fStatus.bits_status.top
#define STATUS_B   fStatus.bits_status.b
#define STATUS_REG fStatus.reg

typedef struct {
	/* general register set */
	Bit32u eax; /* accumulator */
	Bit32u ebx; /* base */
	Bit32u ecx; /* count */
	Bit32u edx; /* data */
	Bit32u ebp; /* base pointer */
	Bit32u esi; /* source index */
	Bit32u edi; /* destination index */
	Bit32u esp; /* stack pointer */

	/* floating point structures */
	FReg_t   		fReg[8];
	FloatTagWord_t 	fTag[8];
	FloatControlReg_t 	fControl;
	FloatStatusReg_t  	fStatus;
	/* currently ignoring the instruction and data pointers */


	/* status and control flags register set */
	flags_reg_t	eflags; /* status and control flags */
	Bit32u	pending_sti;
	Bit32u 	eip;    /* instruction pointer */

	/* user segment register set */
	segment_reg_t  cs; /* code segment register */
	segment_reg_t  ss; /* stack segment register */
	segment_reg_t  ds; /* data segment register */
	segment_reg_t  es; /* extra segment register */
	segment_reg_t  fs; /* extra segment register */
	segment_reg_t  gs; /* extra segment register */

	/* system segment registers */
	segment_reg_t gdtr; /* global descriptor table register */
	segment_reg_t idtr; /* interrupt descriptor table register */
	segment_reg_t ldtr; /* interrupt descriptor table register */
	segment_reg_t tr;   /* task register */
	


	/* pointers to get at registers in bytes and words */
	Bit8u  *gen_reg8[8];
	Bit16u *gen_reg16[8];
	Bit32u *gen_reg32[8];

	Bit16u *base_reg16[8];
	Bit16u *index_reg16[8];

	/* debug registers 0-7 (unimplemented) */
	Bit32u dr0, dr1, dr2, dr3, dr4, dr5, dr6, dr7;

	/* test registers 3-7 (unimplemented) */
	Bit32u tr3, tr4, tr5, tr6, tr7;

	/* Control registers */
	cr0_t  cr0;
} cpu_t;



#ifdef LITTLE_ENDIAN
#define REG8L_OFFSET 0
#define REG8H_OFFSET 1
#define REG16_OFFSET 0
#else /* BIG_ENDIAN */
#define REG8L_OFFSET 3
#define REG8H_OFFSET 2
#define REG16_OFFSET 2
#endif



/* segment register encoding */
#define SEG_REG_ES    0
#define SEG_REG_CS    1
#define SEG_REG_SS    2
#define SEG_REG_DS    3
#define SEG_REG_FS    4
#define SEG_REG_GS    5

/* byte regs */
#define BIT_REG_AL  0
#define BIT_REG_CL  1
#define BIT_REG_DL  2
#define BIT_REG_BL  3
#define BIT_REG_AH  4
#define BIT_REG_CH  5
#define BIT_REG_DH  6
#define BIT_REG_BH  7

/* word regs */
#define BIT_REG_AX 0
#define BIT_REG_CX 1
#define BIT_REG_DX 2
#define BIT_REG_BX 3
#define BIT_REG_SP 4
#define BIT_REG_BP 5
#define BIT_REG_SI 6
#define BIT_REG_DI 7

/* double word regs */
#define BIT_REG_EAX 0
#define BIT_REG_ECX 1
#define BIT_REG_EDX 2
#define BIT_REG_EBX 3
#define BIT_REG_ESP 4
#define BIT_REG_EBP 5
#define BIT_REG_ESI 6
#define BIT_REG_EDI 7

#define NUM_REGS    8


/* global opcode names */
enum Opcode {
  OP_NA = 0,    /* NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) OP,
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFFU
#undef DEFINST
#undef DEFLDST
#undef DEFLINK
  OP_MAX        /* number of opcodes + NA */
};

/* inst -> enum Opcode */
#define OP_ENUM(BITS)   (mask2op[(BITS)])
extern enum Opcode mask2op[];

/* enum Opcode -> description string */
#define OP_NAME(OP)     (op2name[OP])
extern char *op2name[];

/* enum Opcode -> opcode flags */
#define OP_FLAGS(OP)    (op2flags[OP])
extern int op2flags[];


extern void initState(cpu_t *state);
extern void InitPreciseState(cpu_t *state);

extern void DumpState(FILE *stream, cpu_t *state);
extern void DumpFinalStats(FILE *stream);

#endif /* IX86_STATE_H */
