/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/file.h>
#include <bfd.h>

#include "misc.h"
#include "ix86_state.h"
#include "translate.h"
#include "memory.h"

#define DATA_BASE               0x08000000
#define MEM_TABLE_SIZE		0x80000		/* upper mem is not used */
#define MEM_BLOCK_SIZE		0x10000
#define MEM_INDEX(addr) 	((addr) >> 16)
#define MEM_OFFSET(addr)	((addr) & 0xffff)

static unsigned char *instMem;
static int           memcheck;
static unsigned long textStart = 0x08000080;    /* this is where the .init section starts */
static unsigned long textSize;			/* text size in bytes */
static unsigned char *dataMem[MEM_TABLE_SIZE] = {NULL};
static unsigned long dataStart;	       		/* data segment base address */
static unsigned long dataEnd;
static unsigned long dataSize;			/* data size in bytes */
static unsigned long stackStart;		/* starting stack address */

unsigned long        textEnd;			/* end of text */
unsigned long        progEntry;			/* program entry point */
unsigned long        breakPoint = 0;		/* current heap break point */
unsigned long        baseBreak = 0;		/* initial heap break point */
unsigned long minSP = 0xffffffff;	/* min SP seen thus far */

static cpu_t *nonSpecState = NULL;              /* local pointer to the non-speculative machine state */



/*****************************************************************************
 *****************************************************************************/
inline
Bit8u
getIByte(unsigned long addr)
{
	if((addr > (textStart + textSize)) || (addr<textStart))
		fatal("bad instruction byte fetch address 0x%08lx testStart = 0x%08lx, size = 0x%08lx",
		      addr,
		      textStart,
		      textSize);

	return instMem[addr - textStart];

}/*** end getIByte() ***/



/*****************************************************************************
 *****************************************************************************/
static
void
instMemoryAccess(enum MemCmd cmd,	/* Read or Write */
		 unsigned long addr,	/* address to access */
		 unsigned char *p,	/* where to copy to/from */
		 int numBytes)		/* size of the access */
{
	unsigned char *q,
                      *oldp;
	int i;

	q = &instMem[addr - textStart];
	oldp = p;

	switch (cmd) {
	    case Read:
		q += numBytes -1;
		for (i = 0; i < numBytes; i++) {
			*p = *q;
			p++;
			q--;
		}
		break;
	    case Write:
		fatal("write to instruction memory through memoryAccess interface");
	    default:
		fatal("bad cmd to instMemoryAccess");
	}
	p = oldp;

}/*** end instMemAccess() ***/





/*****************************************************************************
  Load and store to main memory.  This function checks all size and
  alignments.  This function can ONLY access data memory, use
  GetInst() to access instruction memory 
 *****************************************************************************/
void
MemoryAccess(enum MemCmd cmd,	/* Read or Write */
	     unsigned long addr,/* address to access */
	     unsigned char *p,	/* where to copy to/from */
	     int numBytes)	/* size of the access */
{
	unsigned char *q;
	unsigned char *oldp;
	int  i;

	/* track the minimum SP */
	minSP = MIN(nonSpecState->esp, minSP);

	/* see if this access is to inst or data memory */
	if((addr >= textStart) && (addr < textEnd)) {
		/* it is an instruction memory access */
		instMemoryAccess(cmd,addr,p,numBytes);
		return;
	}
	
	/* validate the access */
	if (addr < dataStart) {
		error("memory access error: out of bounds, addr 0x%08x\n minSP == 0x%08x nonSpecState->esp = 0x%08x\n dataStart == 0x%08x dataEnd = 0x%08x breakpoint =  0x%08x",
		      addr, minSP, nonSpecState->esp, dataStart, dataEnd, breakPoint);
	}

	/* see if we need to allocate a new section */
	if (!dataMem[MEM_INDEX(addr)]) {
		if (!(dataMem[MEM_INDEX(addr)] = GetCore(MEM_BLOCK_SIZE)))
			error("out of memory");
		/* GetCore() does a bzero() */
	}

	/*
	 * see if this access crosses pages -- if it does, it will
	 * require special handling
	 */
	if (MEM_INDEX(addr) != MEM_INDEX(addr+numBytes-1)) {
		/* special handling required */
		for(i = 0; i < numBytes; i++) {
			if (!dataMem[MEM_INDEX(addr+numBytes-1-i)]) {
				/* we must check the whole range of the access, as things can be
				   arbitrarily aligned */
				if (!(dataMem[MEM_INDEX(addr+numBytes-1-i)] = GetCore(MEM_BLOCK_SIZE)))
					error("out of memory3");
				/* GetCore() does a bzero() */
			}
			q = dataMem[MEM_INDEX(addr+numBytes-1-i)]+ MEM_OFFSET(addr+numBytes-1-i);

			switch(cmd) {
			case Read:
				/* get one byte from the appropriate page */
				*p = *q;
				p++;
				break;
			case Write:
				*q = *p;
				p++;
				break;
			default:
				fatal("bad cmd to memroy access");
			}
		}
		
	} else {
		 /* non-page crossing case */

		q = dataMem[MEM_INDEX(addr)] + MEM_OFFSET(addr);
      
		oldp = p;
		switch (cmd) {
		    case Read:
			q += numBytes -1;
			for (i = 0; i < numBytes; i++) {
				*p = *q;
				p++;
				q--;
			}
			break;
		    case Write:
			p += numBytes - 1;
			for (i = 0; i < numBytes; i++) {
				*q = *p;
				p--;
				q++;
			}
			break;
		    default:
			fatal("bad cmd to memoryAccess");
		}
		p = oldp;
	} /* end of normal, non-page crossing case */
}



/*****************************************************************************
 *****************************************************************************/
unsigned long
CopyToStr(maft_t maf, unsigned long addr, char *s)
{
	char c;
	int i=0;

	do {
		(*maf)(Read, addr++, &c, 1);
		*s = c;
		s++; i++;
		if (i == MAXBUFSIZE)
			error("out of buffer space in handling syscall");
	} while (c != '\0');
	return addr;

}/*** end CopyToStr() ***/





/*****************************************************************************
 *****************************************************************************/
unsigned long
CopyFromStr(maft_t maf, char *s, unsigned long addr)
{
	char c;
	int i=0;

	assert(s);

	do {
		c = *s;
		(*maf)(Write, addr++, &c, 1);
		s++; i++;
		if (i == MAXBUFSIZE)
			error("out of buffer space in handling syscall");
	} while (c != '\0');
	return addr;

}/*** end CopyFromStr() ***/




/****************************************************************************
 ****************************************************************************/
static
void
process_text(bfd *abfd, asection *sect)
{
	unsigned int   i;
	Bit8u          *tempMem;

	if ((tempMem = (Bit8u *)calloc(PAGE_UP(bfd_section_size(abfd, sect)),1)) == NULL)
		fatal("cannot allocate space for %s segment",bfd_section_name(abfd, sect));

	debug("PAGE_UP(bfd_section_size(abfd, sect)) = %lu",
	      PAGE_UP(bfd_section_size(abfd, sect)));

	if (!bfd_get_section_contents(abfd, sect, tempMem, (file_ptr)0,
				      bfd_section_size(abfd, sect)))
		fatal("could not read entire %s section from executable",bfd_section_name(abfd, sect));

	/* now copy from tempMem into correct place in instMem */
	for (i = 0; i < bfd_section_size(abfd, sect); i++)
		instMem[bfd_section_vma(abfd, sect) + i - textStart] = tempMem[i];

	/* now free the temporary memory */
	free(tempMem);

}/*** end processText() ***/




/****************************************************************************
 ****************************************************************************/
static
void
process_data(bfd *abfd, asection *sect)
{
	unsigned long    k;
	unsigned char    *p;
	int      temp = 0;

	if (bfd_get_section_flags(abfd, sect) & SEC_DATA) {
		/* read the section data */
		debug("reading section data from section `%s'",
		      bfd_section_name(abfd, sect));
		if (!(p = calloc(bfd_section_size(abfd, sect), sizeof(unsigned char))))
			fatal("cannot allocate space for section `%s'",
			      bfd_section_name(abfd, sect));
		
		if (!bfd_get_section_contents(abfd, sect, p, (file_ptr)0,
					      bfd_section_size(abfd, sect)))
			fatal("could not read `%s' section from executable",
			      bfd_section_name(abfd, sect));
    
		/* copy it into simulator main memory */
		for (k = 0; k < bfd_section_size(abfd, sect); k++)
			MemoryAccess(Write,
				     bfd_section_vma(abfd, sect) + k,
				     (unsigned char *)&p[k],
				     1);

		free(p);
	} else {
		/* clear the uninitialized data segment in simulator main memory */
		debug("clearing this section's simulator  memory");
		for (k = 0; k < bfd_section_size(abfd, sect); k+=4)
			MemoryAccess(Write,
				     bfd_section_vma(abfd, sect) + k,
				     (unsigned char *)&temp,
				     4);
	}

}/*** end process_data() ***/




/*****************************************************************************
 *****************************************************************************/
void
LoadProgram(int argc, char **argv, char **envp, cpu_t *state)
{
	unsigned long     addr,
	                  sp,
	                  initTarget = 0,
                          finiTarget = 0;
	int               i;
	bfd       *abfd;
	asection  *sect;
	int       gotDataSect = FALSE,
	          gotTextSect = FALSE;

	Boolean   oldVerbose;

	oldVerbose = verbose;
	verbose = FALSE;

	memcheck = 0;

	assert(state);
	nonSpecState = state;

	/* set up a local stack pointer, this is where the argv and envp
	   data is written into program memory */
	/* FIXME -- this value (0xc0000000) is chosen rather arbitrarily....
	   see LINUX kernel hackers guide */
	stackStart = sp = minSP = nonSpecState->esp = 0xc0000000;
	debug("Initial Stack Pointer is at 0x%lx",stackStart);
	
	textEnd = textStart;

	/* copy args and environ into the simulator's MAIN MEMORY, the format
	   in memory is:
	   sp
	   |
	   |
	   V
	   [argc][argv array,NULL][envp array,NULL][argv strings][envp strings] */

	/* write [argc] */
	MemoryAccess(Write, sp, (unsigned char *)&argc, 4);

	/* skip past argv array and NULLs */
	addr = sp + (argc + 2)*4;
	sp += 4;

	/* save space for envp array */
	for (i=0; envp[i] != NULL; i++)
		addr+=4;
	/* skip past envp NULL */
	addr += 4;

	/* fill in the argv array and data */
	for (i = 0; i < argc; i++) {
		/* write the argv array entry */
		MemoryAccess(Write, sp, (unsigned char *)&addr, 4);
		sp += 4;
		/* and the data */
		addr = CopyFromStr(MemoryAccess, argv[i], addr);
	}
	/* terminate argv array */
	i = 0;
	MemoryAccess(Write, sp, (unsigned char *)&i, 4);
	sp += 4;

	/* write envp array and data */
	i = 0;
	while (envp[i]) {
		/* write the envp array entry */
		MemoryAccess(Write, sp, (unsigned char *)&addr, 4);
		sp += 4;
		/* and the data */
		addr = CopyFromStr(MemoryAccess, envp[i], addr);
		i++;
	}
	/* terminate the envp array */
	i = 0;
	MemoryAccess(Write, sp, (unsigned char *)&i, 4);

	/* load the program into memory */
	if (!(abfd = bfd_openr(argv[0], /* default target */NULL)))
		fatal("cannot open executable `%s': bfd_openr failed", argv[0]);
	/* this call is mainly for its side effect of reading in the sections.
	   we follow the traditional behavior of `strings' in that we don't
	   complain if we don't recognize a file to be an object file.  */
	if (!bfd_check_format(abfd, bfd_object)) {
		bfd_close (abfd);
		fatal("cannot open executable `%s': check_format failed", argv[0]);
	}

	debug("processing %d sections...", bfd_count_sections(abfd));

	/* first, walk through all the sections and decide how to allocate memory */
	for (sect=abfd->sections; sect; sect=sect->next) {
		if (!strcmp(bfd_section_name(abfd, sect), ".text")
		    || !strcmp(bfd_section_name(abfd, sect), ".init") 
		    || !strcmp(bfd_section_name(abfd, sect), "__libc_subinit") 
		    || !strcmp(bfd_section_name(abfd, sect), ".fini")) {
			/* must add these sections into the size of the text for this program */
			textStart = MIN(textStart, bfd_section_vma(abfd, sect));
			textEnd =   MAX(textEnd,   bfd_section_vma(abfd, sect)+bfd_section_size(abfd, sect));
			if(oldVerbose) {
				debug("Section %s: %lx bytes",bfd_section_name(abfd, sect),bfd_section_size(abfd, sect));
			}

			if (!strcmp(bfd_section_name(abfd, sect), ".init")) {
				/* this will be the entry point */
				initTarget =  bfd_section_vma(abfd, sect);
			} else if (!strcmp(bfd_section_name(abfd, sect), ".fini")) {
				/* this will be the entry point */
				finiTarget = bfd_section_vma(abfd, sect);
			} else if (!strcmp(bfd_section_name(abfd, sect), ".text")) {
				/* we need to target this section entry after doint the .init stuff */
				progEntry = bfd_section_vma(abfd, sect);
			}
		}
		if (!strcmp(bfd_section_name(abfd, sect),".rodata")
		    || !strcmp(bfd_section_name(abfd, sect), ".ctors")
		    || !strcmp(bfd_section_name(abfd, sect), ".dtors")
		    || !strcmp(bfd_section_name(abfd, sect), ".bss")
		    || !strcmp(bfd_section_name(abfd, sect), ".data")) {
			/* initialize the break point */
			dataStart = MIN(dataStart, bfd_section_vma(abfd, sect));
			dataEnd = MAX(dataEnd, bfd_section_vma(abfd, sect)+bfd_section_size(abfd, sect));

			dataSize += bfd_section_size(abfd, sect);
		}
	}


	/* allocate memory for all the instructions */
	textSize = textEnd-textStart;
	if(oldVerbose) {
		info("text Start is 0x%lx",textStart);
		info("text End   is 0x%lx",textEnd);
		info("total text size is 0x%lx",textSize);
		info("     program entry point = 0x%08lx", progEntry);
		info("PAGE_UP(textSize) = 0x%lx",PAGE_UP(textSize));
		info("data Start is 0x%lx",dataStart);
		info("data End   is 0x%lx",dataEnd);
		info("total data size is 0x%lx",dataSize);
	}
	if ((instMem = (Bit8u *)GetCore(PAGE_UP(textSize))) == NULL)
		fatal("cannot allocate space for `.text' segment");
	bzero(instMem, PAGE_UP(textSize));

	baseBreak = dataEnd;
	breakPoint = PAGE_UP(dataEnd);

	for (sect = abfd->sections; sect; sect = sect->next) {
		debug("processing section `%s', %d bytes @ 0x%08x...",
		      bfd_section_name(abfd, sect), bfd_section_size(abfd, sect),
		      bfd_section_vma(abfd, sect));
		if (!strcmp(bfd_section_name(abfd, sect), ".text") 
		    || !strcmp(bfd_section_name(abfd, sect), "__libc_subinit")
		    || !strcmp(bfd_section_name(abfd, sect), ".fini")
		    || !strcmp(bfd_section_name(abfd, sect), ".init")) {
			/* .text section processing */
			process_text(abfd, sect);
			gotTextSect = TRUE;
		}
		else if (!strcmp(bfd_section_name(abfd, sect), ".rodata")
			 || !strcmp(bfd_section_name(abfd, sect), ".bss")
			 || !strcmp(bfd_section_name(abfd, sect), ".ctors")
			 || !strcmp(bfd_section_name(abfd, sect), ".dtors")
			 || !strcmp(bfd_section_name(abfd, sect), ".data")) {
			/* .data section processing */
			process_data(abfd, sect);
			gotDataSect = TRUE;
		} else if ((!strcmp(bfd_section_name(abfd, sect), ".comment"))
			   || (!strcmp(bfd_section_name(abfd, sect), ".stab"))
			   || (!strcmp(bfd_section_name(abfd, sect), ".stabstr")))
			/* just ignore the .comment sections */;
		else
			debug("encountered unknown section `%s', %d bytes @ 0x%08x",
			      bfd_section_name(abfd, sect), bfd_section_size(abfd, sect),
			      bfd_section_vma(abfd, sect));
	}

	debug("initial breakpoint @ 0x%08x...", baseBreak);

	if (!gotTextSect)
		fatal("Executable is missing a `.text' type section");
	if (!gotDataSect)
		fatal("Executable is missing a `.data' type section");

	memcheck = 1;
	verbose = oldVerbose;

}/*** end LoadProgram() ***/






/*****************************************************************************
 *****************************************************************************/
void
DumpMemory(unsigned long addr, unsigned long len)
{
	unsigned int i;
	int          k;

	for (i = 0; i < len; i++) {
		MemoryAccess(Read,
			     addr+i,
			     (unsigned char *)&k,
			     1);
		fprintf(stderr, "0x%08lx: %02x\n", addr+i, k);
	}

}/*** end DumpMemory() ***/






/*****************************************************************************
 * print memory usage stats
 *****************************************************************************/
#define IN_K(N)		(((N)+1023) / 1024)
void
MemStats(FILE *stream)
{
	long sum = 0;

	fprintf(stream, "MemStats: total stack used: %8ldk\n",
		(sum += IN_K(stackStart-minSP), IN_K(stackStart-minSP)));
	fprintf(stream, "MemStats: total heap used:  %8ldk\n",
		(sum += IN_K(breakPoint-baseBreak), IN_K(breakPoint-baseBreak)));
	fprintf(stream, "MemStats: total .data used: %8ldk\n",
		(sum += IN_K(baseBreak-DATA_BASE), IN_K(baseBreak-DATA_BASE)));
	fprintf(stream, "MemStats: total memory:     %8ldk\n", sum);

}/*** end MemStats() ***/

#undef IN_K(N)
