/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef MISC_H
#define MISC_H

#include <stdio.h>
#include <sys/types.h>		/* ushort is here */
#include <setjmp.h>

#include "types.h"
#include "errorRoutines.h"

/* missing external decls */
#ifdef sparc
FILE *	fopen(const char *__filename, const char *__type);
char *	fgets(char *__s, int __n, FILE *__stream);
int	fprintf(FILE *__stream, const char *__format, ...);
int	fclose(FILE *__stream);
int	fgetc(FILE *__stream);
int	fputc(int __c, FILE *__stream);
int	fputs(const char *__s, FILE *__stream);
size_t	fread(void *__ptr, size_t __size,
	      size_t __nitems, FILE *__stream);
void	fflush(FILE *stream);
int	fscanf(FILE *__stream, const char *__format, ...);
char *	gets(char *__s);
int	printf(const char *__format, ...);
int	puts(const char *__s);
int	scanf(const char *__format, ...);
int	sscanf(const char *__s, const char *__format, ...);
int	vfprintf(FILE *__stream, const char *__format, ...);
int	vprintf(const char *__format, ...);
int	vsprintf(char *__s, const char *__format, ...);
void *	malloc(size_t size);
void *	calloc(size_t nelt, size_t el_size);
void *	realloc(void *ptr, size_t size);
void	free(void *ptr);
time_t	time(time_t *);
void	bzero(void *, int);
char *	sbrk(int incr);
#endif /* sparc */




#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* all options */
#define ALL_OPTIONS "W:Vv:m:"

extern enum SimType {
	Functional,
	CacheSimulation,
	TLBSimulation,
	SuperScalar,
	RUUSimulation,
	EventTrace,
	SCCSimulation
} simType;

#ifdef MICRO_OPS
extern int micro_ops_this_cycle;
#endif
extern long long         numInsts; 	/* speculative instructions executed */
extern long long         nnumInsts; 	/* non-speculative instructions executed */
extern long long         numDecodeCacheHits;
extern long long         numCycles; 	/* issue stats */
extern int               dumpStats; 	/* set this for intermediate stats */
extern time_t            startTime; 	/* start time in seconds */
extern jmp_buf           exitBuf; 	/* jump to this at exit time */
extern unsigned long     verbInst; 	/* when to get verbose (instruction number) */
extern unsigned long     verbAddr; 	/* when to get verbose (instruction Address) */
extern unsigned long     verbTime; 	/* when to get verbose (cycle number) */
extern char              *execName; 	/* program name */
extern unsigned long     lastSyscall; 	/* for debugging condor problems */

extern int               verbose;       /* debug flag */
extern int               syscallVerbose;/* dump debug info only during system calls */

/* This array defines a look-up table for the even parity-ness
   of an 8bit quantity, for optimal assignment of the parity bit
   in the EFLAGS register
  */
extern int parity_lookup[256];


/* various useful macros */
#ifndef MAX
#define MAX(a, b)    (((a) < (b)) ? (b) : (a))
#endif
#ifndef MIN
#define MIN(a, b)    (((a) < (b)) ? (a) : (b))
#endif

#define info(fmt, args...)	\
    _info(__FILE__, __FUNCTION__, __LINE__, fmt, ## args)

extern void _info(char *file, char *func, int line, char *fmt, ...);

#ifndef NDEBUG

#define debug(fmt, args...)	\
    do {                        \
        if (verbose)            \
            _debug(__FILE__, __FUNCTION__, __LINE__, fmt, ## args); \
    } while(0)

extern void _debug(char *file, char *func, int line, char *fmt, ...);

#else

#define debug(fmt, args...)

#endif /* NDEBUG */


/* copy a string to a new storage allocation */
extern char *strdup(const char *s);


/* allocate some core, this memory has overhead no larger than a page
   in size and it cannot be released. the storage is returned cleared */
extern void *GetCore(unsigned long size);

/* return string describing elapsed time, passed in SEC in seconds */
extern char *ElapsedTime(long sec);

/* getopt() interfaces */
extern int    optInd;
extern int    optErr;
extern char   *optArg;
extern void   NewOpt(void);
extern char   GetOpt(int argc, char **argv, char *fmt);

extern Bit16u zeroExtend8_16( Bit8u byte);
extern Bit32u zeroExtend8_32( Bit8u byte);
extern Bit32u zeroExtend16_32( Bit16u byte);
extern Bit16u signExtend8_16( Bit8u byte);
extern Bit32u signExtend8_32( Bit8u byte);
extern Bit32u signExtend16_32( Bit16u word);

#define SX_B2W(BYTE)      (signExtend8_16(BYTE))
#define SX_B2D(BYTE)      (signExtend8_32(BYTE))
#define SX_W2D(WORD)      (signExtend16_32(WORD))
#define ZX_B2W(BYTE)      (zeroExtend8_16(BYTE))
#define ZX_B2D(BYTE)      (zeroExtend8_32(BYTE))
#define ZX_W2D(WORD)      (zeroExtend16_32(WORD))


#endif /*** end ifndef MISC_H ***/
