/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

/***********************************************************************************
 * INCLUDES
 ***********************************************************************************/
#include <stdio.h>

#include "misc.h"
#include "operands.h"
#include "non-spec.defines"

/***********************************************************************************
 * DEFINES
 ***********************************************************************************/
#ifdef MICRO_OPS
#define MOPP micro_ops_this_cycle++;
#else
#define MOPP
#endif


/***********************************************************************************
 * FORWARDS
 ***********************************************************************************/
static Bit32u getByteSIB(iInst_t inst, cpu_t *state);
static Bit32u computeRMAdd(iInst_t inst, cpu_t *state);





/***********************************************************************************
 ***********************************************************************************/
void
MyMemoryAccess(enum MemCmd cmd,     /* Read or Write */
	       unsigned long addr,  /* address to access */
	       unsigned char *p,    /* where to copy to/from */
	       int numBytes)        /* size of the access */
{
	MemoryAccess(cmd,addr,p,numBytes);

	MOPP;

	if(verbose) {
		fprintf(stderr,"            Mem Access: %s to addr:0x%08lx %d bytes\n",
			(cmd == Read) ? " Read":"Write", addr, numBytes);
		switch(numBytes) {
		    case 1:
			fprintf(stderr,"\t\tvalue %s = 0x%02x.\n",
				(cmd == Read)?"read":"written", *((Bit8u *) p));
			break;
		    case 2:
			fprintf(stderr,"\t\tvalue %s = 0x%04x.\n",
				(cmd == Read)?"read":"written", *((Bit16u *) p));
			break;
		    case 4:
			fprintf(stderr,"\t\tvalue %s = 0x%08lx.\n",
				(cmd == Read)?"read":"written", *((Bit32u *) p));
			break;
		    case 8:
			fprintf(stderr,"\t\tvalue %s = %f.\n",
				(cmd == Read)?"read":"written",
				*((double *)p));
			break;
		}
	}

}/*** end MyMemoryAccess() ***/





/**************************************************************************
 **************************************************************************/
static
Bit32u
getByteSIB(iInst_t inst, cpu_t *state)
{
	assert(state);
	assert(inst.base >= 0);
	assert(inst.base < 8);
	assert(inst.mod >= 0);
	assert(inst.mod < 4);
	assert(inst.ss >= 0);
	assert(inst.ss < 4);
	assert(inst.index >= 0);
	assert(inst.index < 8);
	assert(inst.rm == 4);

#define D32     inst.displacement
#define D8      (SX_B2D((Bit8u)(inst.displacement & 0x000000ff)))
#define S_I      ((inst.index != 4) ? ((*(state->gen_reg32[inst.index])) * (1 << inst.ss)) : (0))
#define BASE    (*(state->gen_reg32[inst.base]))
	switch(inst.mod) {
	    case 0: /* DS/SS:[base + (scaled index)] */
		if (inst.base == 5)
			return (D32 + S_I);
		else
			return (BASE + S_I);
		break;
	    case 1: /* DS/SS:[base + (scaled index) + d8] */
		return ( BASE + S_I + D8 );
		break;
	    case 2: /* DS/SS:[base + (scaled index) + d32] */
		return ( BASE + S_I + D32 );
		break;
	    default: fatal("bad mod value in getByteSIB");
	}
#undef D32
#undef D8
#undef S_I
#undef BASE
}

/***************************************************************************
 ***************************************************************************/
static
Bit32u
computeRMAdd(iInst_t inst, cpu_t *state)
{
	assert(state);
	assert(inst.mod >= 0);
	assert(inst.mod < 4);
	assert(inst.rm >= 0);
	assert(inst.rm < 8);
	assert(inst.rmreg >= 0);
	assert(inst.rmreg < 8);

#define REG (*(state->gen_reg32[inst.rm]))
#define D8  (SX_B2D((Bit8u)(inst.displacement & 0x000000ff)))
#define D32 (inst.displacement)

	switch(inst.mod) {
	    case 0: /* DS:[reg] */
		if (inst.rm == 4)
			/* sib byte present */
			return getByteSIB(inst, state);
		
		if (inst.rm == 5)
			return (D32);
		
		return (REG);

	    case 1: /* DS[reg + d8] */
		if (inst.rm == 4)
			/* sib byte present */
			return(getByteSIB(inst, state));
		
		return (REG + D8);

	    case 2: /* DS:[reg + d32] */
		if (inst.rm == 4)
			/* sib byte present */
			return (getByteSIB(inst, state));

		return (REG + D32);
		
	    default: fatal("illegal mod value in computeRMadd");
	}
#undef REG
#undef D8
#undef D32

}/*** end computeRMAdd() ***/


/**************************************************************************
 **************************************************************************/
Bit8u
getByteOp(iInst_t inst, char type, cpu_t *state)
{
	Bit8u  byte;
	Bit32u addr;
	
	switch(type) {
	    case 'A': return (*(state->gen_reg8[BIT_REG_AL]));

	    case 'H': return (*(state->gen_reg8[BIT_REG_AH]));

	    case 'C': return (*(state->gen_reg8[BIT_REG_CL]));

	    case 'l': return (*(state->gen_reg8[BIT_REG_DL]));

	    case 'G': return (*(state->gen_reg8[inst.rmreg]));

	    case '=': /* must look at low three bits of opcode to determine register nubmer */
		    return (*(state->gen_reg8[inst.opcodeByte & 0x07]));

	    case 'I': return (Bit8u) inst.immediate;

	    case 'O': return (Bit8u) inst.offset;

	    case 'R': assert(inst.addSize != WORD);
		    if (inst.mod != 3) {
			    addr = computeRMAdd(inst, state);
			    MyMemoryAccess(Read, addr, (unsigned char *)&byte,1);
		    } else
			    byte = *(state->gen_reg8[inst.rm]);
		    return byte;

	    case 'M': /* use simple offset 16 or 32 bits determined by addSize */
		    assert(inst.addSize != WORD);
		    MyMemoryAccess(Read, inst.immediate, (unsigned char *)&byte,1);
		    return byte;

	    case 'd': MyMemoryAccess(Read, *(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&byte, 1);
		    return byte;
		    
	    case 'q': MyMemoryAccess(Read, *(state->gen_reg32[BIT_REG_ESI]), (unsigned char *)&byte, 1);
		    return byte;

	    default:
		    fatal("bad operand format in getByteOp: %c\n",type);
	}

}/*** end getByteOp() ***/




/***************************************************************************
 ***************************************************************************/
Bit16u
getWordOp(iInst_t inst, char type, cpu_t *state)
{
	Bit16u word;
	Bit32u addr;

	switch(type) {
	    case 'd':
		assert(0);
		MyMemoryAccess(Read, *(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&word, 2);
		return word;

	    case 'M': /* use simple offset 16 or 32 bits determined by addSize */
		assert(inst.addSize != WORD);
		MyMemoryAccess(Read, inst.immediate, (unsigned char *)&word,2);
		return word;

	    case 'I': return (Bit16u) inst.immediate;

	    case 'O': return (Bit16u) inst.offset;

	    case 'A': return (*(state->gen_reg16[BIT_REG_AX]));

	    case 'G': return (*(state->gen_reg16[inst.rmreg]));

	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Read, addr, (unsigned char *)&word,2);
		} else
			word = *(state->gen_reg16[inst.rm]);
		return word;

	    case '=': /* must look at low three bits of opcode to determine register nubmer */
		return (*(state->gen_reg16[inst.opcodeByte & 0x07]));

	    default: 
		fatal("bad operand format in getWordOp: %c\n",type);
	}

}/*** end getWordOp() ***/




/***************************************************************************
 ***************************************************************************/
Bit32u
getDoubOp(iInst_t inst, char type, cpu_t *state)
{
	Bit32u doub;
	Bit32u addr;

	switch(type) {
	    case 'A': return *(state->gen_reg32[BIT_REG_EAX]);

	    case 'C': return *(state->gen_reg32[BIT_REG_ECX]);

	    case 'l': return *(state->gen_reg32[BIT_REG_EDX]);

	    case 'D': return *(state->gen_reg32[BIT_REG_EDI]);

	    case 'G': return *(state->gen_reg32[inst.rmreg]);

	    case '=': /* must look at low three bits of opcode to determine register nubmer */
		    return (*(state->gen_reg32[inst.opcodeByte & 0x07]));

	    case 'I': return (Bit32u) inst.immediate;
		    
	    case 'O': return (Bit32u) inst.offset;

	    case 'R':
		    if (inst.mod != 3) {
			    addr = computeRMAdd(inst, state);
			    MyMemoryAccess(Read, addr, (unsigned char *)&doub,4);
		    } else
			    doub = *(state->gen_reg32[inst.rm]);
		    return doub;

	    case 'M':
		    /* use simple offset 16 or 32 bits determined by addSize */
		    MyMemoryAccess(Read, inst.immediate, (unsigned char *)&doub,4);
		    return doub;
		    
	    case 'd':
		    MyMemoryAccess(Read, *(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&doub, 4);
		    return doub;
		    
	    case 'q':
		    MyMemoryAccess(Read, *(state->gen_reg32[BIT_REG_ESI]), (unsigned char *)&doub, 4);
		    return doub;
		    
	    case 'Q': return *(state->gen_reg32[BIT_REG_ESI]);
		    
	    default:
		    fatal("bad operand format in getDoubOp: %c\n",type);
	}

}/*** end getDoubOp() ***/





/****************************************************************************
 ****************************************************************************/
Bit64u
getQuadOp(iInst_t inst, char type, cpu_t *state)
{
	Bit64u quad;
	Bit32u addr;

	switch(type) {
	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Read, addr, (unsigned char *)&quad,8);
		} else
			assert(0);
		return quad;

	    default:
		fatal("bad operand format in getDoubOp: %c\n",type);
	}

}/*** end getQuadOp() ***/




/***************************************************************************
 ***************************************************************************/
Bit32u
getDoubAddr(iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;

	switch(type) {
	    case 'R':
		if (inst.mod == 3)
			fatal("bad mod value for getDoubAddr");
		addr = computeRMAdd(inst, state);
		if(verbose)
			fprintf(stderr,"  getDoubaddr: 0x%08lx\n",addr);
		return addr;
	    default:
		fatal("bad address format in getDoubAddr: %c\n",type);
	}

}/*** end getDoubAddr() ***/



/**************************************************************************
 **************************************************************************/
void
storeByteOp(Bit8u res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;

	switch(type) {
	    case 'A':
		*(state->gen_reg8[BIT_REG_AL]) = res;
		break;

	    case 'G':
		assert(inst.rmreg >= 0);
		assert(inst.rmreg < 8);
		*(state->gen_reg8[inst.rmreg]) = res;
		break;

	    case '=':
		/* must look at low three bits of opcode to determine register number */
		*(state->gen_reg8[inst.opcodeByte & 0x07]) = res;
		break;

	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 1);
		} else
			*(state->gen_reg8[inst.rm]) = res;
		break;

	    case 'M':
		assert(inst.addSize != WORD);
		MyMemoryAccess(Write, inst.immediate, (unsigned char *)&res,1);
		break;

	    case 'd':
		MyMemoryAccess(Write,*(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&res, 1);
		break;

	    case 'q':
		MyMemoryAccess(Write,*(state->gen_reg32[BIT_REG_ESI]), (unsigned char *)&res, 1);
		break;

	default:
		fatal("bad type to storeByteOp():%c",type);
	}

}/*** end storeByteOp() ***/




/***************************************************************************
 ***************************************************************************/
void
storeWordOp(Bit16u res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;

	switch(type) {
	    case 'd':
		MyMemoryAccess(Write,*(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&res, 2);
		break;

	    case 'A':
		*(state->gen_reg16[BIT_REG_AX]) = res;
		break;

	    case 'l':
		*(state->gen_reg16[BIT_REG_DX]) = res;
		break;

	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 2);
		} else
			*(state->gen_reg16[inst.rm]) = res;
		break;

	    case 'M':
		assert(inst.addSize != WORD);
		MyMemoryAccess(Write,inst.immediate,(unsigned char *)&res, 2);
		break;

	    case 'G':
		assert(inst.rmreg >= 0);
		assert(inst.rmreg < 8);
		*(state->gen_reg16[inst.rmreg]) = res;
		break;

	    case '=':
		/* must look at low three bits of opcode to determine register nubmer */
		*(state->gen_reg16[inst.opcodeByte & 0x07]) = res;
		break;

	    default:
		fatal("bad type to storeWordOp():%c",type);
	}

}/*** end storeWordOp() ***/

/**************************************************************************
 **************************************************************************/
void
storeDoubOp(Bit32u res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;

	switch(type) {
	    case 'A':
		*(state->gen_reg32[BIT_REG_EAX]) = res;
		break;

	    case 'C':
		*(state->gen_reg32[BIT_REG_ECX]) = res;
		break;

	    case 'l':
		*(state->gen_reg32[BIT_REG_EDX]) = res;
		break;

	    case 'D':
		*(state->gen_reg32[BIT_REG_EDI]) = res;
		break;

	    case 'Q':
		*(state->gen_reg32[BIT_REG_ESI]) = res;
		break;
		
	    case 'G':
		assert(inst.rmreg >= 0);
		assert(inst.rmreg < 8);
		*(state->gen_reg32[inst.rmreg]) = res;
		break;

	    case '=':
		/* must look at low three bits of opcode to determine register nubmer */
		*(state->gen_reg32[inst.opcodeByte & 0x07]) = res;
		break;

	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 4);
		} else
			*(state->gen_reg32[inst.rm]) = res;
		break;

	    case 'M':
		assert(inst.addSize != WORD);
		MyMemoryAccess(Write,inst.immediate,(unsigned char *)&res, 4);
		break;

	    case 'd':
		MyMemoryAccess(Write,*(state->gen_reg32[BIT_REG_EDI]), (unsigned char *)&res, 4);
		break;

	    case 'q':
		MyMemoryAccess(Write,*(state->gen_reg32[BIT_REG_ESI]), (unsigned char *)&res, 4);
		break;

	    default:
		fatal("bad type to storeDoubOp():%c",type);
	}

}/*** end storeDoubOp() ***/



/**************************************************************************
 **************************************************************************/
void
storeQuadOp(Bit64s res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;

	switch(type) {
	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 8);
		} else
			fatal("don't know how to store quad to reg file");
		break;

	    default:
		fatal("bad type to storeQuadOp():%c",type);
	}

}/*** end storeQuadOp() **/





/*************************************************************************
 *************************************************************************/
Float32
getFloat32Op(iInst_t inst, char type, cpu_t *state) 
{
	Float32 ff;
	int     i;
	Bit32u  addr;

	switch(type) {
	    case 'R':
		assert(inst.rm >= 0);
		assert(inst.rm <  8);
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Read, addr, (unsigned char *)&ff,4);
		} else {
			i = state->STATUS_TOP + inst.rm;
			assert(i >= 0);
			assert(i < 15);
			if (i > 7)
				i -= 8;
			if(verbose)
				fprintf(stderr," getting value from freg %d(actual)/%d(requested):",
					i,inst.rm);
			ff = (state->fReg[i]);
		}
		if(verbose)
			fprintf(stderr,"returning %f from getFloat32Op\n",ff);
		return ff;

	    default: fatal("bad type to getFloat32Op:'%c'",type);
	}

}/*** end getFloat32Op() ***/




/***************************************************************************
 ***************************************************************************/
Float64
getFloat64Op(iInst_t inst, char type, cpu_t *state) 
{
	Float64 ff;
	Bit32u addr;
	int    i;

	switch(type) {
	    case '|':
		i = state->STATUS_TOP + (inst.opcodeByte & 0x0007);
		assert(i >= 0);
		assert(i < 15);
		if (i > 7)
			i -= 8;
		ff = (state->fReg[i]);
		if(verbose)
			fprintf(stderr,"   getFloat64Op: returning %f from freg %d(actual)/%d(requested)\n",
				ff,i,inst.opcodeByte&0x07);
		return ff;

	    case 'R':
		assert(inst.rm >= 0);
		assert(inst.rm <  8);
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Read, addr, (unsigned char *)&ff,8);
		} else {
			i = state->STATUS_TOP + inst.rm;
			assert(i >= 0);
			assert(i < 15);
			if (i > 7)
				i -= 8;
			if(verbose)
				fprintf(stderr,"    getFloat64Op: getting value from freg %d(actual)/%d(requested):",i,inst.rm);
			ff = (state->fReg[i]);
		}
		if(verbose)
			fprintf(stderr,"    getFloat64Op: returning %f from getFloat64Op\n",
				(double)ff);
		return ff;

	    default:
		fatal("bad type to getFloat64Op:'%c'",type);
	}

}/*** end getFloat64Op() ***/



/**************************************************************************
 **************************************************************************/
void
storeFloat32Op(Float32 res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;
	int i;

	switch(type) {
	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 4);
		} else {
			assert(inst.rm >=0);
			assert(inst.rm <8);
			i = state->STATUS_TOP + inst.rm;
			assert(i >= 0);
			assert(i < 15);
			if (i > 7)
				i -= 8;
			(state->fReg[i]) = (double)res;
			(state->fTag[i]) = FPU_TAG_VALID;
			if(verbose)
				fprintf(stderr,"     storing value %f to freg %d(actual)/%d(requested)\n",
					res, i,inst.rm);
		}
		break;

	    default: fatal("bad type to storeFloat32Op:'%c'",type);
	}

}/*** end storeFloat32Op() ***/



/***************************************************************************
 ***************************************************************************/
void
storeFloat64Op(Float64 res, iInst_t inst, char type, cpu_t *state)
{
	Bit32u addr;
	int i;

	switch(type) {
	    case '|':
		i = state->STATUS_TOP + (inst.opcodeByte & 0x0007);
		assert(i >= 0);
		assert(i < 15);
		if (i > 7)
			i -= 8;
		(state->fReg[i]) = res;
		if(verbose)
			fprintf(stderr,"   storeFlaot64Op: storing %f to freg #%d(actual)/%d(requested)\n",
				res,i,inst.opcodeByte&0x07);
		break;

	    case 'R':
		if (inst.mod != 3) {
			addr = computeRMAdd(inst, state);
			MyMemoryAccess(Write, addr, (unsigned char *)&res, 8);
		} else {
			assert(inst.rm >=0);
			assert(inst.rm <8);
			i = state->STATUS_TOP + inst.rm;
			assert(i >= 0);
			assert(i < 15);
			if (i > 7)
				i -= 8;
			(state->fReg[i]) = res;
			(state->fTag[i]) = FPU_TAG_VALID;
			if(verbose)
				fprintf(stderr,"   storeFloat64Op: storing %f to freg #%d(actual)/%d(requested)\n",
					res,i,inst.rm);
		}
		break;

	    default: fatal("bad type to storeFloat64Op:'%c'",type);
	}

}/*** end storeFloatOp64() ***/




/**************************************************************************
 **************************************************************************/
void
push_fop32(Float32 value, cpu_t *state)
{
	/* decrement stack pointer */
	if (state->STATUS_TOP == 0)
		state->STATUS_TOP = 0x7;
	else
		state->STATUS_TOP--;
	
	/* FIXME: check for overflow */
	if (state->CONTROL_IM == 0)
		assert((state->fTag[state->STATUS_TOP]) == FPU_TAG_EMPTY);
	
	/* push the value onto the stack */
	state->fReg[state->STATUS_TOP] = (double)value;

	if (verbose)
		fprintf(stderr,"    push_fop32: %f into slot %d(actual)\n",
			value,state->STATUS_TOP);
	
	/* indicate that the stack position is valid */
	state->fTag[state->STATUS_TOP] = FPU_TAG_VALID;

}/*** end push_fop32() ***/

void
push_fop64(Float64 value, cpu_t *state)
{
	/* decrement stack pointer */
	if (state->STATUS_TOP == 0)
		state->STATUS_TOP = 0x7;
	else
		state->STATUS_TOP--;
	
	/* FIXME: check for overflow */
	if (state->CONTROL_IM == 0)
		assert((state->fTag[state->STATUS_TOP]) == FPU_TAG_EMPTY);
	
	/* push the value onto the stack */
	state->fReg[state->STATUS_TOP] = value;
	
	if (verbose)
		fprintf(stderr,"    push_fop64: %f into slot %d(actual)\n",
			value,state->STATUS_TOP);
	
	/* indicate that the stack position is valid */
	state->fTag[state->STATUS_TOP] = FPU_TAG_VALID;

}/*** end push_fop64() ***/



/*************************************************************************
 *************************************************************************/
Float64
fpu_stack(int depth, cpu_t *state)
{
	int entry;

	if(state->STATUS_TOP + depth > 7)
		entry = state->STATUS_TOP + depth - 8;
	else
		entry = state->STATUS_TOP + depth;
	
	assert(entry >= 0);
	assert(entry <  8);

	if ((state->fTag[entry]) == FPU_TAG_EMPTY)
		fatal(" requested stack entry: %d(requested)/%d(actual) not valid",depth,entry);
	
	if(verbose)
	  fprintf(stderr,"   fpu_stack: returning %f from stack slot %d(actual)/%d(requested)\n",
                  state->fReg[entry],entry,depth);

	return (state->fReg[entry]);

}/*** end fpu_stack() ***/





/*************************************************************************
 *************************************************************************/
void
pop_fop(cpu_t *state)
{
	/* FIXME: check for underflow */
	if (state->CONTROL_IM == 0)
		assert((state->fTag[state->STATUS_TOP]) != FPU_TAG_EMPTY);

	/* invalidate top of stack */
	state->fTag[state->STATUS_TOP] = FPU_TAG_EMPTY;

	if (verbose)
		fprintf(stderr,"   pop_fop: popping slot %d(actual)\n",state->STATUS_TOP);

	/* increment stack pointer */
	if (state->STATUS_TOP == 0x07)
		state->STATUS_TOP = 0x00;
	else
		state->STATUS_TOP++;

}/*** end pop_fop() ***/

