/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

/*************************************************************
      syscall.c - interpret and execute Linux system calls
 *************************************************************/

/***************************************
 INCLUDE FILES
 ***************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#include <syscall.h>
#include <errno.h>
#ifndef sparc
#include <sys/ioctl.h>
#else
#include <sys/ttold.h>
#endif
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <sys/file.h> /* includes <sys/fcntlcom.h> */
#include <sys/dirent.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <setjmp.h>
#include <termio.h>
#include <sys/dir.h>
#include <sys/times.h>

#include "misc.h"
#include "ix86_state.h"
#include "memory.h"
#include "syscall.h"
#include "errorRoutines.h"
#include "translate.h"

/* decls missing in the system header files */
extern int ioctl(int fd, int req, caddr_t arg);
extern int getrlimit(int resource, struct rlimit *rlp);
extern int setrlimit(int resource, struct rlimit *rlp);
extern int lstat(char *path,struct stat *buf);
extern int rename(char *path1, char *path2);
extern int getrusage(int who, struct rusage *rusage);
extern int gettimeofday(struct timeval *tp, struct timezone *tzp);
extern int utimes(char *file, struct timeval *tvp);


/****************************************************************
 FUNCTION PROTOTYPES
 ****************************************************************/
static void Syscall(maft_t maf, iInst_t inst, cpu_t *state);
static void swapBytes(int num_bytes, unsigned char *where);

static void translateCharL2S(maft_t maf, unsigned long LINUX_char, unsigned char *sun_char);
static void translateCharS2L(maft_t maf, unsigned char *sun_char, unsigned long LINUX_char);

static void translateIntL2S(maft_t maf, unsigned long LINUX_int, unsigned char *sun_int);
static void translateIntS2L(maft_t maf, unsigned char *sun_int, unsigned long LINUX_int);

static void translateRusageL2S(maft_t maf, unsigned long LINUX_ru, unsigned char *l_rue);
static void translateRusageS2L(maft_t maf, unsigned char *s_rue, unsigned long LINUX_ru);

static void translateRlimitL2S(maft_t maf, unsigned long LINUX_rl, unsigned char *s);
static void translateRlimitS2L(maft_t maf, unsigned char *s, unsigned long LINUX_rl);

static void translateStatL2S(maft_t maf, unsigned long LINUX_sb, unsigned char *s_s);
static void translateStatS2L(maft_t maf, unsigned char *s_s, unsigned long LINUX_sb);

static void translateDES2L(maft_t maf, unsigned char *s_de, unsigned long sim_de);

static void translateTermiosL2S(maft_t maf, unsigned long LINUX_t, unsigned char *s_t);
static void translateTermiosS2L(maft_t maf, unsigned char * s_t, unsigned long LINUX_t);

static void translateTermioL2S(maft_t maf, unsigned long LINUX_t, unsigned char *s_t);
static void translateTermioS2L(maft_t maf, unsigned char * s_t, unsigned long LINUX_t);

static void translateWinSizeL2S(maft_t maf, unsigned long LINUX_ws, unsigned char *s_ws);
static void translateWinSizeS2L(maft_t maf, unsigned char *s_ws, unsigned long LINUX_ws);

static int           translateOpenFlags(unsigned long Lflags);
static int           translateFcntlCommand(unsigned long command);
static unsigned long translateFcntlArgument(unsigned long argument, int local_command);
static int           translateWho(long lwho);
static int           translateResource(unsigned long resource);
static int           translateMode(unsigned long mode);

static void translateTimevalL2S(maft_t maf, unsigned long Ltimeval, unsigned char *st_v);
static void translateTimevalS2L(maft_t maf, unsigned char * st_v, unsigned long Ltimeval);

static void translateTimezoneL2S(maft_t maf, unsigned long Ltimezone, unsigned char *st_z);
static void translateTimezoneS2L(maft_t maf, unsigned char *st_z, unsigned long Ltimezone);

static void translatePipeFDS2L(maft_t maf, unsigned char *sun_fd, unsigned long Lfds);

static void translateTMSS2L(maft_t maf, unsigned char *sun_tms, unsigned long Ltms);
static void translateTimeTS2L(maft_t maf, unsigned char *sun_t, unsigned long Lt);

static void seterrnum(int err);

static void insertFD(unsigned long fd, char *filename, int flags, unsigned long mode);
static void deleteFD(unsigned long fd);
static char *getFName(unsigned long fd);
static DIR  *getDP(unsigned long fd, char *filename);
static void insertDP(unsigned long fd, char *filename, DIR *dirp);

static unsigned long errNumAddr = 0;


typedef struct cpCache {
  unsigned long  fd;
  char           *filename;
  DIR            *dirp;
  int            flags;
  unsigned long  mode;
  struct cpCache *next;
} cpCache_t;

static cpCache_t *dpCacheHead = NULL;


/******************************************************************
 INTERRUPT HANDLER
 ******************************************************************/
void
interrupt(Bit8u number, iInst_t inst, cpu_t *state)
{
  switch(number){
      case 0x80:
	/* system call */
	Syscall(MemoryAccess,inst, state);
	break;

      case 0x03:
	/* trap to debugger */
	fatal("got trap to debugger!!!\n");
	break;

      case 0x04:
	/* overflow trap */
	fatal("got overflow trap");
	break;

      default:
	/* unknown interrupt vector */
	fatal("unknown interrupt requested:%d",number);
	break;
  }
}/*** end interrupt() ***/


/******************************************************************
 SYSTEM CALL ROUTINES
 ******************************************************************/
static
void
Syscall(maft_t maf, iInst_t inst, cpu_t *state)
{
  unsigned long syscode = state->eax;

  lastSyscall = syscode;

  switch (syscode) {
      /***************************************
       PERSONALITY -- 0x88
       ***************************************/
      case LINUX_SYS_personality:
        {
          /* verifies that the environment is legal Linux environment --
             see kernel/exec_domain.c: sys_personality() */
          state->eax = 0;

          if(verbose || syscallVerbose)
		  fprintf(stderr, "syscall: sy	s_personality: returning 0\n");

        }
        break;

      /***************************************
       ALARM
       ***************************************/
      case LINUX_SYS_alarm:
         {
           unsigned long 	alarm = state->ebx;

           if(verbose || syscallVerbose)
		   fprintf(stderr, "sy	scall: alarm - alarm value = 0x%08lx\n",alarm);

           state->eax = 0;
         }
        break;

      /***************************************
       BRK -- 0x2d
       ***************************************/
      case LINUX_SYS_brk:
         {
           /* returns brk value */
           unsigned long 	brk = state->ebx,
	                        esp = state->esp;

           if(verbose || syscallVerbose)
		   fprintf(stderr, "syscall: sys_brk: brk value == 0x%08lx\n",brk);

           if (brk < textEnd) {
		   state->eax = breakPoint;
		   if(verbose || syscallVerbose)
			   fprintf(stderr, "\t\t\tretu	rning 0x%08lx\n",breakPoint);
		   return;
           }

           /* round the new heap pointer to the its page boundary */
           /* check whether heap area has merged with stack area */
           brk = PAGE_UP(brk);
           if ((brk >= baseBreak) && (brk < esp)) {
		   breakPoint = brk;
		   state->eax = brk;
		   if(verbose || syscallVerbose)
			   fprintf(stderr,"\t\t\tretur	ning 0x%08lx\n",breakPoint);
		   return;
           }
           else
             fatal("SYS_brk failed");
        }
        break;

      /***************************************
       NEWFSSTAT -- 0x6c
       ***************************************/
      case LINUX_SYS_newfstat:
        {
          unsigned long ret_val,
			fd 	 = state->ebx,
			LstatBuf = state->ecx;
          unsigned char 	sBuf[sizeof(struct stat)];

          translateStatL2S(maf, LstatBuf, sBuf);

          /* do the call -- returns 0 if successful */
          ret_val = fstat((signed)fd, (struct stat *)sBuf);

          if(verbose || syscallVerbose) {
		  fprintf(stderr, "syscall new	fstat: fd = 0x%lx\n",fd);
		  if ((signed)ret_val < 0)
			  fprintf(stderr,"\t\t\tunsucc	essful ret_value = 0x%lx\n",ret_val);
          }

          if (ret_val != 0)
		  seterrnum(errno);

          translateStatS2L(maf, sBuf, LstatBuf);

          state->eax = ret_val;
        }
        break;

      /***************************************
       EXIT -- 0x01
       ***************************************/
      case LINUX_SYS_exit:
        {
          if(verbose || syscallVerbose)
		  fprintf(stderr, "syscall exit: ret	urn value = 0x%lx\n",state->eax+1);

          longjmp(exitBuf, /* exitcode */(signed)state->eax+1);
        }
        break;

      /***************************************
       READ
       ***************************************/
      case LINUX_SYS_read:
        {
          unsigned long fd     = state->ebx,
                        nBytes = state->edx,
                        lBuf   = state->ecx;
	  int           numRead,
	                i;
          char          *buf;

          if (!(buf = (char *)calloc(nBytes, sizeof(char))))
		  fatal("out of memory in SYS_r	ead");

          numRead = read((signed)fd , buf, nBytes);

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_read: nbytes = 0	x%08lx, num read = %d, fd# = 0x%08lx\n",
			  nBytes,
			  numRead,
			  fd);

          if (numRead == -1)
		  seterrnum(errno);

          for (i = 0; i < numRead; i++)
		  (*maf)(Write, lBuf+i, (unsigned char *)&buf[i], 1);

          state->eax = numRead;

          free(buf);
        }
        break;

      /***************************************
       WRITE -- 0x04
       ***************************************/
      case LINUX_SYS_write:
        {
          unsigned long 	i,
                        numWritten,
			fd     = state->ebx,
			nBytes = state->edx,
          		Lbuf   = state->ecx;
          char          *buf;

          if (!(buf = (char *)calloc(nBytes, sizeof(char))))
		  fatal("out of memory in SYS_read");

          for (i=0; i < nBytes; i++)
		  (*maf)(Read, Lbuf+i, (unsigned char *)&buf[i], 1);

          /* do the call -- returns the number of bytes written */
          numWritten = write((signed)fd, buf, nBytes);

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_write: nbytes = 0x%08lx, num written = 0x%08lx, fd# = 0x%08lx\n",
			  nBytes,
			  numWritten,
			  fd);
          if (numWritten != nBytes)
		  fatal("SYS_wri	te failed");

          state->eax = numWritten;

          free(buf);
        }
        break;

      /***************************************
       NEW(L)STAT -- 0x6a and 0x6b
       ***************************************/
      case LINUX_SYS_newstat:
      case LINUX_SYS_newlstat:
        {
          unsigned long  ret_val,
		 fName = state->ebx,
		 LsBuf = state->ecx;
          char   buf[MAXBUFSIZE];		/* for the file name string */
          unsigned char  sBuf[sizeof(struct stat)];	/* for the stat buffer */

          /* read file name string from simulator memory */
          CopyToStr(maf, fName, buf);

          if(verbose || syscallVerbose)
            fprintf(stderr,"syscall_%s: file name = '%s'\n\n",
                      (syscode ==  LINUX_SYS_newstat)? "newstat":"newlstat",
		    buf);

          /* translate stat buffer from LINUX to SUN */
          translateStatL2S(maf, LsBuf, sBuf);

          /* do the call -- returns 0 is successful */
          if (syscode ==  LINUX_SYS_newstat)
		  ret_val = stat(buf, (struct stat *)sBuf);
          else
		  ret_val = lstat(buf, (struct stat *)sBuf);

          if(verbose || syscallVerbose) {
		  if (ret_val != 0)
			  fprintf(stderr,"\t\t\tfailed: errno = 0x%04x\n",errno);
		  else
			  fprintf(stderr,"\t\t\treturn value = 0x%08lx\n",ret_val);
          }

          if (ret_val != 0)
		  seterrnum(errno);

          /* translate stat buffer from SUN to LINUX */
          translateStatS2L(maf, sBuf, LsBuf);

          state->eax = ret_val;
        }
        break;

      /***************************************
       OPEN
       ***************************************/
      case LINUX_SYS_open:
        {
          char		filename[MAXBUFSIZE];
          unsigned long flags 		= state->ecx,
                        mode  		= state->edx,
			fName 		= state->ebx;
	  int           translatedFlags = 0;
	  int           fd;

          CopyToStr(maf, fName, filename);

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_open: file name = '%s'; flags = 0x%08lx(LINUX);\n",
			  filename,
			  flags);

          translatedFlags = translateOpenFlags(flags);

          /* FIXME: this assumes that the mode layout is the same -- may not always
                    be the case - we should check */

          /* do the call -- returns fd or -1 */
          fd = open(filename, translatedFlags, (signed)mode);

          if(verbose || syscallVerbose) {
		  fprintf(stderr,"syscall_open: file name = '%s'; flags = 0x%08lx(LINUX)/0x%08x(SUN);\n",
			  filename,
			  flags,
			  translatedFlags);
		  fprintf(stderr,"\t\t\t\tmode = 0x%08lx(LINUX)/0x%08lx(SUN); fd (ret val) = %d\n",
			  mode,
			  mode,
			  fd);
		  if (fd == -1)
			  fprintf(stderr,"\t\t\t\tfailed -- errno = 0x%04x\n",
				  errno);
          }

          if (fd == -1)
		  seterrnum(errno);
          else
		  insertFD((unsigned)fd, filename, translatedFlags, mode);

          state->eax = fd;
        }
        break;

      /***************************************
       CLOSE
       ***************************************/
      case LINUX_SYS_close:
        {
          unsigned long	fd = state->ebx;
	  int           ret_val;
          char          *filename;
          DIR           *dirp = NULL;

          /* don't close stdin, stdout, or stderr */
          if ((fd == 0) ||
	      (fd == 1) ||
	      (fd == 2))
		  return;

          /* verify that we see an open file with this file descriptor */
          if ((filename = getFName(fd)) != NULL) {
		  dirp = getDP(fd, filename);
		  
		  if(dirp != NULL)
			  ret_val = closedir(dirp);
          }

          ret_val = close((signed)fd );

          if(verbose || syscallVerbose) {
		  if (filename != NULL)
			  fprintf(stderr,"syscall_c	lose: fd = 0x%08lx; file name = '%s'; directory? %s\n",
				  fd,
				  filename,((dirp==NULL)?"NO":"YES"));
		  else
			  fprintf(stderr,"syscall_close: fd = 0x%08lx; UNOPENED!\n", fd);

		  if (ret_val == -1)
			  fprintf(stderr,"\t\t\t	\tfailed -- errno = 0x%04x\n",errno);
          }

          if (ret_val == -1)
		  seterrnum(errno);

          /* free the descriptor cache entry for this fd */
          if (filename != NULL)
		  deleteFD(fd);

          state->eax = ret_val;
        }
        break;

      /***************************************
       IOCTL
       ***************************************/
      case LINUX_SYS_ioctl:
        {
#define NUM_IOCTL_BYTES         128
          int            ret_val;
	  unsigned long  fd        = state->ebx,
		         request   = state->ecx,
		         LINUX_arg = state->edx,
          	         local_req = 0;
          unsigned char  *local_arg = NULL;
          struct termios sun_termios;
          struct termio  sun_termio;
          struct winsize sun_winSize;
          int            sun_int;
          char           sun_char;


/* convert to local ioctl request values */
          switch (request) {
		  /*
		     case LINUX_TIOCCONS: 	case LINUX_TIOCPKT: 	case LINUX_TIOCNOTTY: 	case LINUX_TIOCSETD:
		     case LINUX_TIOCGETD: 	case LINUX_FIONBIO:	case LINUX_FIONREAD: 	case LINUX_TIOCGSERIAL:
		     case LINUX_TIOCSSERIAL: 	case LINUX_TCSBRKP:	case LINUX_TIOCTTYGSTRUCT:
		     case LINUX_FIONCLEX: 	case LINUX_FIOCLEX: 	case LINUX_FIOASYNC: case LINUX_TIOCSERCONFIG:
		     case LINUX_TIOCSERGWILD: 	case LINUX_TIOCSERSWILD: 	case LINUX_TIOCGLCKTRMIOS:
		     case LINUX_TIOCSLCKTRMIOS: 	case LINUX_TIOCSERGSTRUCT: 	case LINUX_TIOCSERGETLSR:
		     case LINUX_TIOCSERGETMULTI: case LINUX_TIOCSERSETMULTI: 	case LINUX_TIOCLINUX:
		     */
	      case LINUX_TCGETS:
		  local_req = TCGETS;
		  local_arg = (unsigned char *)&sun_termios;
		  translateTermiosL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETS:
		  local_req = TCSETS;
		  local_arg = (unsigned char *)&sun_termios;
		  translateTermiosL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETSW:
		  local_req = TCSETSW;
		  local_arg = (unsigned char *)&sun_termios;
		  translateTermiosL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETSF:
		  local_req = TCSETSF;
		  local_arg = (unsigned char *)&sun_termios;
		  translateTermiosL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCGETA:
		  local_req = TCGETA;
		  local_arg = (unsigned char *)&sun_termio;
		  translateTermioL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETA:
		  local_req = TCSETA;
		  local_arg = (unsigned char *)&sun_termio;
		  translateTermioL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETAW:
		  local_req = TCSETAW;
		  local_arg = (unsigned char *)&sun_termio;
		  translateTermioL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSETAF:
		  local_req = TCSETAF;
		  local_arg = (unsigned char *)&sun_termio;
		  translateTermioL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCGPGRP:
		  local_req = TIOCGPGRP;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCSPGRP:
		  local_req = TIOCSPGRP;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCOUTQ:
		  local_req = TIOCOUTQ;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCSTI:
		  local_req = TIOCSTI;
		  local_arg = (unsigned char *)&sun_char;
		  translateCharL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCGWINSZ:
		  local_req = TIOCGWINSZ;
		  local_arg = (unsigned char *)&sun_winSize;
		  translateWinSizeL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCSWINSZ:
		  local_req = TIOCSWINSZ;
		  local_arg = (unsigned char *)&sun_winSize;
		  translateWinSizeL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCMGET:
		  local_req = TIOCMGET;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCMBIS:
		  local_req = TIOCMBIS;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCMBIC:
		  local_req = TIOCMBIC;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCMSET:
		  local_req = TIOCMSET;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCGSOFTCAR:
		  local_req = TIOCGSOFTCAR;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TIOCSSOFTCAR:local_req = TIOCSSOFTCAR;
		  local_arg = (unsigned char *)&sun_int;
		  translateIntL2S(maf, LINUX_arg,local_arg);
		  break;
	      case LINUX_TCSBRK:
		  local_req = TCSBRK;	/* arg is an int */
		  break;
	      case LINUX_TCXONC:
		  local_req = TCXONC;	/* arg is an int */
		  break;
	      case LINUX_TCFLSH:
		  local_req = TCFLSH; 	/* arg is an int */
		  break;
	      case LINUX_TIOCEXCL:
		  local_req = TIOCEXCL;	/* arg is ignored */
		  break;
	      case LINUX_TIOCNXCL:
		  local_req = TIOCNXCL;	/* arg is ignored */
		  break;
	      case LINUX_TIOCSCTTY:
		  local_req = TIOCSCTTY; /* arg is an int */
		  break;

	      default:
		  fatal("syscall: SYS_ioctl: not supported fd=0x%08lx, req=0x%08x",
			fd, request);
          }

          /* do the call */
          if (local_arg != NULL)
		  ret_val = ioctl((signed)fd, (signed)local_req, (caddr_t)&local_arg[0]);
          else
		  ret_val = ioctl((signed)fd, (signed)local_req, (caddr_t)LINUX_arg);

          if((request == LINUX_TCGETS) && (fd == 0) && (ret_val == -1)) {
		  assert(local_req==TCGETS);

		  ret_val = ioctl(1, (signed)local_req, (caddr_t)&local_arg[0]);

		  if(verbose || syscallVerbose) {
			  fprintf(stderr,"\t\t\t weird TCGETS kludge....\n");
			  fprintf(stderr,"\t\t\t   ret_val using stdout = %d\n",ret_val);
		  }
	  }

          switch (request) {
              case LINUX_TCGETS:	
              case LINUX_TCSETS:	
              case LINUX_TCSETSW:
		  translateTermiosS2L(maf, local_arg,LINUX_arg);
		  break;
              case LINUX_TCSETSF:	
              case LINUX_TCGETA:	
              case LINUX_TCSETA:	
              case LINUX_TCSETAW:	
              case LINUX_TCSETAF:
		  translateTermioS2L(maf, local_arg,LINUX_arg);
		  break;
              case LINUX_TIOCMGET:
              case LINUX_TIOCMBIS:
              case LINUX_TIOCMBIC:
              case LINUX_TIOCMSET:
              case LINUX_TIOCGSOFTCAR:
              case LINUX_TIOCSSOFTCAR:
              case LINUX_TIOCGPGRP:
              case LINUX_TIOCSPGRP:
              case LINUX_TIOCOUTQ:
		  /* arg is an int * */
		  translateIntS2L(maf, local_arg,LINUX_arg);
		  break;
              case LINUX_TIOCSTI:
		  translateCharS2L(maf, local_arg,LINUX_arg);
		  break;
              case LINUX_TIOCGWINSZ:
              case LINUX_TIOCSWINSZ:
		  translateWinSizeS2L(maf, local_arg,LINUX_arg);
		  break;

              case LINUX_TCSBRK:
              case LINUX_TCXONC:
              case LINUX_TCFLSH:
              case LINUX_TIOCSCTTY:
		  /* arg is an int */
		  break;
              case LINUX_TIOCEXCL:
              case LINUX_TIOCNXCL:
		  /* arg is ignored */
		  break;
              default:
		  fatal("syscall: SYS_ioctl(2): not supported fd=0x%08lx, req=0x%08x",
			fd, request);
          }

          if(verbose || syscallVerbose) {
		  fprintf(stderr,"syscall_ioctl: fd = 0x%08lx;\trequest = 0x%08lx(LINUX)/0x%08lx(SUN);\n\t\targp = 0x%08lx(LINUX)/0x%08x(SUN)\n",
			  fd,
			  request,
			  local_req,
			  LINUX_arg,
			  *local_arg);
		  if (ret_val == -1)
			  fprintf(stderr,"\t\t\tfailed -- errno = 0x%x\n",errno);
          }
			
          if (ret_val == -1)
		  seterrnum(errno);

          state->eax = ret_val;
        }
	break;

      /***************************************
       FCNTL
       ***************************************/
      case LINUX_SYS_fcntl:
        {
          unsigned long command  = state->ecx,
                        argument = state->edx,
                        fd       = state->ebx,
			local_argument;
	  int           local_command;
	  int           ret_val;

          /* translate the command from LINUX to SUN */
          local_command = translateFcntlCommand(command);

          /* translate the argument from LINUX to SUN */
          local_argument = translateFcntlArgument(argument,local_command);

          /* do the call */
          ret_val = fcntl((signed)fd, local_command, local_argument);

          if(verbose || syscallVerbose) {
		  fprintf(stderr,"syscall_fcntl: fd = 0x%08lx;\tcommand = 0x%08lx(LINUX)/0x%08x(SUN);\n",
			  fd, command, local_command);
		  fprintf(stderr,"\t\t\t\targument = 0x%08lx(LINUX)/0x%08lx(SUN); ret_val = %d\n",
			  argument, local_argument, ret_val);
		  if(ret_val != 0)
			  fprintf(stderr,"\t\t\t\tfailed: errno == 0x%04x\n",errno);
          }

          if (ret_val == -1)
		  seterrnum(errno);

          state->eax = ret_val;
        }
        break;

      /***************************************
       READDIR
       ***************************************/
      case LINUX_SYS_readdir:
        {
            unsigned long  fd 	= state->ebx,
			   count 	= state->edx,
			   sim_de 	= state->ecx;
	    char	   *filename = NULL;
            DIR		   *dirp = NULL;
            unsigned char  *sun_de;
            int		   olderrno;

            if((filename = getFName(fd)) == NULL)
		    fatal("tried to readdir on an unopened fd value");

            dirp = getDP(fd,filename);
            if(dirp == NULL) {
		    dirp = opendir(filename);
		    if (dirp == NULL)
			    fatal("syscall: opendir failed\n");
		    else
			    insertDP(fd, filename, dirp);
            }

            olderrno = errno;

            sun_de = (unsigned char *)readdir(dirp);

            /* translate returned dirent * to LINUX */
            if (sun_de != NULL)
		    translateDES2L(maf, sun_de,sim_de);

            if(verbose || syscallVerbose) {
		    fprintf(stderr,"syscall_readdir: fd = 0x%08lx; dir name = '%s'; count = 0x%08lx ret_val = 0x%08lx\n",
			    fd, filename, count, (sun_de == NULL)?0:sim_de);
		    if(sun_de == NULL)
			    fprintf(stderr, "\t\t\treturned NULL: errno = 0x%04x; olderrno = 0x%04x\n",errno,olderrno);
		    else
			    fprintf(stderr, "\t\t\treturned string = '%s'\n",((struct dirent *)sun_de)->d_name);
            }

            if(sun_de == NULL) {
		    seterrnum(errno);
		    state->eax = 0;
            } else
		    state->eax = sim_de;
        }
        break;

      /***************************************
       GET[.]*ID
       ***************************************/
      case LINUX_SYS_getpgid:
      case LINUX_SYS_getppid:
      case LINUX_SYS_getpid:
      case LINUX_SYS_getgid:
      case LINUX_SYS_getuid:
      case LINUX_SYS_geteuid:
      case LINUX_SYS_getegid:
        {
          char      string[100];
          int       ret_val;

          switch(syscode) {
            case LINUX_SYS_getpgid:
		  fatal("getpgid called");
		  /*ret_val = getpgid(); */
		  strcpy(string,"syscall_getpgid");
		  break;
            case LINUX_SYS_getppid:
		  ret_val = getppid();
		  strcpy(string,"syscall_getppid");
		  break;
            case LINUX_SYS_getpid:
		  ret_val = getpid();
		  strcpy(string,"syscall_getpid");
		  break;
            case LINUX_SYS_getuid:
		  ret_val = getuid();
		  strcpy(string,"syscall_getuid");
		  break;
            case LINUX_SYS_getgid:
		  ret_val = getgid();
		  strcpy(string,"syscall_getgid");
		  break;
            case LINUX_SYS_geteuid:
		  ret_val = geteuid();
		  strcpy(string,"syscall_geteuid");
		  break;
            case LINUX_SYS_getegid:
		  ret_val = getegid();
		  strcpy(string,"syscall_getegid");
		  break;
            default:
		  fatal("what the hell????");
          }

          if(verbose || syscallVerbose)
		  fprintf(stderr, "%s: returning %d\n",string, ret_val);

          if(ret_val == -1)
		  seterrnum(errno);

          state->eax = ret_val;
        }
        break;

      /***************************************
       CREAT
       ***************************************/
      case LINUX_SYS_creat:
        {
          char 		filename[MAXBUFSIZE];
          unsigned long mode  = state->ecx,
                	fName = state->ebx;
	  int           fd;

          /* I think this is ok, but not needed yet...*/
          fatal("creat() called");

          CopyToStr(maf, fName, filename);

          /* do the call */
          fd = creat(filename, mode);

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_creat -- filename = '%s', fd = %d; mode = 0x%08lx\n",
			  filename,
			  fd,
			  mode);

          if (fd ==  -1)
		  seterrnum(errno);
          else
		  insertFD((unsigned)fd, filename, 0, mode);

          state->eax = fd;
        }
        break;

      /***************************************
       RENAME
       ***************************************/
      case LINUX_SYS_rename:
        {
		int            result;
		unsigned long  old_FName = state->ebx,
                               new_FName = state->ecx;
		char           old_filename[MAXBUFSIZE],
		               new_filename[MAXBUFSIZE];

		CopyToStr(maf, old_FName, old_filename);
		CopyToStr(maf, new_FName, new_filename);

		/* do the call */
		result = rename(old_filename, new_filename);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_rename -- old_filename = '%s', new_filename = '%s'; return value = %d\n",
				old_filename,
				new_filename,
				result);

		if (result ==  -1)
			seterrnum(errno);

		state->eax = result;
        }
        break;

      /***************************************
       RLIMIT stuff
       ***************************************/
      case LINUX_SYS_setrlimit:
      case LINUX_SYS_getrlimit:
        {
		struct rlimit   sun_rl;
		unsigned long	lrlimit  = state->ecx, /* pointer to rlimit structure */
		                resource = state->ebx, /* resource number */
		                result;
		int             local_resource;

		local_resource = translateResource(resource);

		if (lrlimit != 0) {
			/* translate the 			LINUX rlimit struct to SUN */
			translateRlimitL2S(maf, lrlimit, (unsigned char *)&sun_rl);

			if(verbose || syscallVerbose)
				fprintf(stderr,"syscall_get	/setrlimit -- rlim struct before call -> soft = 0x%08x(SUN), hard = 0x%08x(SUN)\n",
					sun_rl.rlim_cur,
					sun_rl.rlim_max);
		}

		if (syscode == LINUX_SYS_getrlimit)
			result = getrlimit(local_resource, &sun_rl);
		else
			result = setrlimit(local_resource, &sun_rl);

		if(verbose || syscallVerbose)
			fprintf(stderr, "syscall_%s -- resource = 0x%08lx(LINUX)/0x%08x(SUN); return value = %lx\n",
				(syscode == LINUX_SYS_getrlimit) ? "getrlimit":"setrlimit",
				resource,
				local_resource,
				result);

		if(result != 0)
			seterrnum(errno);

		if (lrlimit != 0) {
			if(verbose || syscallVerbose)
		    
			  /* translate the SUN rlimit struct to LINUX */
			  translateRlimitS2L(maf, (unsigned char *)&sun_rl,lrlimit);

			if(verbose || syscallVerbose)
				fprintf(stderr,"sysca	ll_get/setrlimit -- rlim struct after call -> soft = 0x%08x(SUN), hard = 0x%08x(SUN)\n",
				  sun_rl.rlim_cur,
					sun_rl.rlim_max);
		}

		state->eax = result;
        }
        break;

      /***************************************
       SIGNAL stuff
       ***************************************/
      case LINUX_SYS_sigprocmask:
      case LINUX_SYS_sigsetmask:
      case LINUX_SYS_sigaction:
        {
		/* just return 0 -- the simulator will catch all signals anyways */
		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_%s -- returning 0\n",
				(syscode == LINUX_SYS_sigsetmask)? "sigsetmask": "sigaction");
		state->eax = 0;
        }
        break;

      /***************************************
       RUSAGE
       ***************************************/
      case LINUX_SYS_getrusage:
        {
          unsigned long lRusage = state->ecx,
                        lwho    = state->ebx,
                        result;
	  int           local_who;
          struct rusage rusage;

          local_who = translateWho((signed)lwho);

          if ( lRusage != 0)
		  /* translate the LINUX rusage struct to SUN */
		  translateRusageL2S(maf, lRusage,(unsigned char *)&rusage);

          result = getrusage(local_who, &rusage);

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_getrusage -- result = %lx\n", result);

          if (lRusage != 0)
		  translateRusageS2L(maf, (unsigned char *)&rusage, lRusage);

          state->eax = result;
        }
        break;

      /***************************************
       ACCESS
       ***************************************/
      case LINUX_SYS_access:
        {
		char 		fName[MAXBUFSIZE];
		unsigned long	lpath  = state->ebx,
		                mode   = state->ecx;
		int             local_mode;
		int             result;
		
		CopyToStr(maf, lpath, fName);

		local_mode = translateMode(mode);

		result = access(fName, local_mode);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_access -- file = '%s', mode = 0x%08lx(LINUX)/0x%08x(SUN); result = %d\n",
				fName,mode,
				local_mode,
				result);

		if (result == -1)
			seterrnum(errno);

		/* FIXME: do we really need to do this? */
		if (lpath != 0)
			CopyFromStr(maf, fName, lpath);

		free(fName);

		state->eax = result;
        }
        break;

      /***************************************
       LSEEK
       ***************************************/
      case LINUX_SYS_lseek:
        {
		unsigned long	fd     = state->ebx,
	                        offset = state->ecx,
		                dir    = state->edx;
		int             result;

		/*
		 * FIXME: note that the whence constants SEEK_SET,CUR
		 * and END are the same in LINUX and sun
		 * currently. this may change.
		 */
		result = lseek((signed)fd, (signed)offset, (signed)dir);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_lseek: fd = 0x%08lx, offset = 0x%08lx, dir = 0x%08lx, result = %d\n",
				fd,offset,dir,result);
		if (result == -1)
			seterrnum(errno);

		state->eax = result;
        }
        break;

      /***************************************
       UNLINK
       ***************************************/
      case LINUX_SYS_unlink:
        {
		unsigned long fName = state->ebx;
                int           result;
		char buf[MAXBUFSIZE];

		CopyToStr(maf, fName, buf);

		result  = unlink(buf);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_unlink: file name = '%s', result = %d\n",
				buf,result);

		/* FIXME: we should delete the appropriate fd from the fd cache */

		if (result == -1)
			seterrnum(errno);

        }
        break;

      /***************************************
       PIPE
       ***************************************/
      case LINUX_SYS_pipe:
        {
          unsigned long fds = state->eax;
	  int           result;
          int		fd[2];
          

          /* read the fd from simulator memory */
          /* no need...just pass in fd[0] and fd[1] -- the system call fills the fd's */

          result = pipe(fd);

          if (result == -1)
		  error("LINUX_SYS_pipe failed");

          if(verbose || syscallVerbose)
		  fprintf(stderr,"syscall_pipe: pipe1 = 0x%04x, pipe2 = 0x%04x; result = %d\n",fd[0],fd[1],result);

          translatePipeFDS2L(maf,(unsigned char *)&fd[0],fds);

          state->eax = fd[0];
          state->eax = fd[1];
          state->eax = result;
        }
        break;

      /***************************************
       GETTIMEOFDAY()
       ***************************************/
      case LINUX_SYS_gettimeofday:
        {
		unsigned long 	Ltimeval  = state->ebx,
		                Ltimezone = state->edx;
		int             result;
		struct timeval	v,
		                *tv = &v;
		struct timezone	z,
		                *tz = &z;

		fatal("can't handle gettimeofday syscalls!");
		assert(0);

		translateTimevalL2S(maf,Ltimeval,(unsigned char *)tv);
		translateTimezoneL2S(maf,Ltimezone,(unsigned char *)tz);

		result = gettimeofday(tv, tz);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_gettimeofday: result = %d\n",result);

		if (result == -1)
			seterrnum(errno);

		translateTimevalS2L(maf,(unsigned char *)tv,Ltimeval);
		translateTimezoneS2L(maf,(unsigned char *)tz,Ltimezone);

		state->eax = result;
        }
        break;

      /***************************************
       TIMES
       ***************************************/
      case LINUX_SYS_times:
        {
		unsigned long  Ltimes = state->ebx;
		int            result;
		struct tms     t,
                               *Stms = &t;

		result = times(Stms);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_times:\n\t\tStms.tms_utime=0x%lx, Stmts.tms_stime=0x%lx, nStms.tms_cutime=0x%lx, Stmts.tms_cstime=0x%lx\n\t\tresult = %d\n",
				t.tms_utime, t.tms_stime,t.tms_cutime, t.tms_cstime,result);

		if (result == -1)
			seterrnum(errno);

		translateTMSS2L(maf,(unsigned char *)Stms,Ltimes);
		
		state->eax = result;
        }
        break;

      /***************************************
       TIME
       ***************************************/
      case LINUX_SYS_time:
        {
		unsigned long Ltime = state->ebx;
		int           result;
		time_t        t;

		result = time(&t);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_time: result = %d\n",result);

		if (result == -1)
			seterrnum(errno);

		translateTimeTS2L(maf,(unsigned char *)&result,Ltime);

		state->eax = result;
        }
        break;

      /***************************************
       FORK
       ***************************************/
      case LINUX_SYS_fork:
        {
		fatal("can't handle forks!");
		assert(0);
        }
        break;

      /***************************************
       CHMOD
       ***************************************/
      case LINUX_SYS_chmod:
	{
		char          buf[MAXBUFSIZE];
		unsigned long fName = state->ebx,
	                      mode  = state->ecx;
		int           result;

		CopyToStr(maf, fName, buf);

		result = chmod(buf, mode );

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_chmod: filename = '%s', mode = 0x%08lx, result = %d\n",
				buf, mode, result);

		if (result == -1)
			seterrnum(errno);

		state->eax = result;	  
	}
	break;


      /***************************************
       CHOWN
       ***************************************/
      case LINUX_SYS_chown:
	{
		char          buf[MAXBUFSIZE];
		unsigned long fName = state->ebx,
                              owner = state->ecx,
                              group = state->edx;
                int           result;

		CopyToStr(maf, fName, buf);

		result = chown(buf, owner, group);

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_chown: filename = '%s', owner = 0x%08lx, group = 0x%08lx, result = %d\n",
				buf, owner, group, result);

		if (result == -1)
			seterrnum(errno);

		state->eax = result;
	}
	break;

      /***************************************
       UTIME
       ***************************************/ 
      case LINUX_SYS_utime:
	{
		unsigned long i,
                              tval,
                              tout,
                              fName;
                int           result;
		char          buf[MAXBUFSIZE],
		              *p;

		fatal("can't handle utime syscalls!");
		assert(0);

#if(0)
		CopyToStr(maf, fName, buf);

		if (tval == 0)
			result = utimes(buf, NULL);
		else {
			struct timeval tval[2];
	      
			/* FIXME -- do the appropriate translation here!!! */
			state->eax = 0;
			return;

			p = (char *)&tval[0];
			for (i = 0; i < 2*sizeof(struct timeval); i++)
				(*maf)(Read, tout+i, p+i, 1);
			
			result = utimes(buf, tval);
		}

		if(verbose || syscallVerbose)
			fprintf(stderr,"syscall_utimes: filename = '%s', ?????? , result = %d\n",
				buf, result);

		if (result == -1)
			seterrnum(errno);

		state->eax = result;
#endif
	}
	break;

      /***************************************
       UNIMPLEMENTED
       ***************************************/ 
      case LINUX_SYS_settimeofday:
      case LINUX_SYS_stime:
      case LINUX_SYS_ftime:

      case LINUX_SYS_sigpending:
      case LINUX_SYS_sigsuspend:
      case LINUX_SYS_signal:
      case LINUX_SYS_sigreturn:

      case LINUX_SYS_setup:
      case LINUX_SYS_waitpid:
      case LINUX_SYS_link:
      case LINUX_SYS_execve:
      case LINUX_SYS_chdir:
      case LINUX_SYS_mknod:
      case LINUX_SYS_break:
      case LINUX_SYS_stat:
      case LINUX_SYS_mount:
      case LINUX_SYS_umount:
      case LINUX_SYS_ptrace:
      case LINUX_SYS_fstat:
      case LINUX_SYS_pause:
      case LINUX_SYS_stty:
      case LINUX_SYS_gtty:
      case LINUX_SYS_nice:
      case LINUX_SYS_sync:
      case LINUX_SYS_kill:
      case LINUX_SYS_mkdir:
      case LINUX_SYS_rmdir:
      case LINUX_SYS_dup:
      case LINUX_SYS_prof:

      case LINUX_SYS_setuid:
      case LINUX_SYS_setgid:
      case LINUX_SYS_setpgid:
      case LINUX_SYS_setsid:
      case LINUX_SYS_setreuid:
      case LINUX_SYS_setregid:
      case LINUX_SYS_setfsuid:
      case LINUX_SYS_setfsgid:

      case LINUX_SYS_acct:
      case LINUX_SYS_phys:
      case LINUX_SYS_lock:
      case LINUX_SYS_mpx:
      case LINUX_SYS_ulimit:
      case LINUX_SYS_uname:
      case LINUX_SYS_umask:
      case LINUX_SYS_chroot:
      case LINUX_SYS_ustat:
      case LINUX_SYS_dup2:
      case LINUX_SYS_getpgrp:
      case LINUX_SYS_sgetmask:
      case LINUX_SYS_sethostname:
      case LINUX_SYS_getgroups:
      case LINUX_SYS_setgroups:
      case LINUX_SYS_select:
      case LINUX_SYS_symlink:
      case LINUX_SYS_lstat:
      case LINUX_SYS_readlink:
      case LINUX_SYS_uselib:
      case LINUX_SYS_swapon:
      case LINUX_SYS_reboot:
      case LINUX_SYS_mmap:
      case LINUX_SYS_munmap:
      case LINUX_SYS_truncate:
      case LINUX_SYS_ftruncate:
      case LINUX_SYS_fchmod:
      case LINUX_SYS_fchown:
      case LINUX_SYS_getpriority:
      case LINUX_SYS_setpriority:
      case LINUX_SYS_profil:
      case LINUX_SYS_statfs:
      case LINUX_SYS_fstatfs:
      case LINUX_SYS_ioperm:
      case LINUX_SYS_socketcall:
      case LINUX_SYS_syslog:
      case LINUX_SYS_getitimer:
      case LINUX_SYS_setitimer:
      case LINUX_SYS_iopl:
      case LINUX_SYS_vhangup:
      case LINUX_SYS_idle:
      case LINUX_SYS_vm86:
      case LINUX_SYS_wait4:
      case LINUX_SYS_swapoff:
      case LINUX_SYS_sysinfo:
      case LINUX_SYS_ipc:
      case LINUX_SYS_fsync:
      case LINUX_SYS_setdomainname:
      case LINUX_SYS_newuname:
      case LINUX_SYS_modify_ldt:
      case LINUX_SYS_adjtimex:
      case LINUX_SYS_mprotect:
      case LINUX_SYS_create_module:
      case LINUX_SYS_init_module:
      case LINUX_SYS_delete_module:
      case LINUX_SYS_get_kernel_syms:
      case LINUX_SYS_quotactl:
      case LINUX_SYS_fchdir:
      case LINUX_SYS_bdflush:
      case LINUX_SYS_sysfs:
      case LINUX_SYS_bllseek:




      default:
	fatal("invalid/unimplemented system call code encountered, syscode %d",
	      syscode);

  }/*** end of big case statement ***/

}/*** end Syscall() ***/





/******************************************************************
 delete file in the cache
 ******************************************************************/
static
void
deleteFD(unsigned long fd)
{
	cpCache_t *last = NULL,
                  *dp   = dpCacheHead;

	if(verbose || syscallVerbose)
		fprintf(stderr,"deleteFD() deleting fd = 0x%08lx\n",fd);

	while(dp != NULL) {
		if (dp->fd == fd) {
			if(last == NULL) {
				assert(dp == dpCacheHead);
				dpCacheHead = dp->next;
			}
			else
				last->next = dp->next;

			if(verbose || syscallVerbose)
				fprintf(stderr,"deleteFD() deleting fd = 0x%08lx; filename = '%s'\n",fd,dp->filename);

			free(dp->filename);
			free(dp);

			return;
		}

		last = dp;
		dp = dp->next;
	}

	fatal("deleteFD() called for non-existant fd\n");

}/*** end deleteFD() ***/



/******************************************************************
 insert a new file in the cache
 ******************************************************************/
static
void
insertFD(unsigned long fd, char *filename, int flags, unsigned long mode)
{
	cpCache_t *dp = dpCacheHead;

	while(dp != NULL) {
		if (dp->fd == fd)
			fatal("inserting a fd multiple times");
		dp = dp->next;
	}

	if((dp = (cpCache_t *)malloc(sizeof(struct cpCache))) == NULL)
		fatal("malloc failed in insertFD()");

	if ((dp->filename = (char *)malloc(1+strlen(filename) * sizeof(char))) == NULL)
		fatal("malloc failed in insertFD():filename");

	strcpy(dp->filename,filename);
	dp->fd 	        = fd;
	dp->flags 	= flags;
	dp->mode 	= mode;
	dp->next 	= dpCacheHead;
	dpCacheHead 	= dp;
	dp->dirp 	= NULL;

	if(verbose || syscallVerbose)
		fprintf(stderr,"insertFD: fd = 0x%08lx; filename = '%s', flags = 0x%08x; mode = 0x%08lx\n",
			fd,
			filename,
			flags,
			mode);

}/*** end insertFD() ***/


/*************************************************************************
 search the cache of DIR pointers for a given fd and return the filename
 *************************************************************************/
static
char *
getFName(unsigned long fd)
{
	cpCache_t *dp = dpCacheHead;

	while(dp != NULL) {
		if (dp->fd == fd)
			return dp->filename;
		dp = dp->next;
	}

	return NULL;

}/*** end getFName() ***/

/******************************************************************
 search the CACHE of DIR pointers for a given fd and filename
 ******************************************************************/
static
DIR *
getDP(unsigned long fd, char *filename)
{
	cpCache_t *dp = dpCacheHead;

	while (dp != NULL) {
		if((dp->fd == fd) && (!strcmp(dp->filename,filename)))
			return dp->dirp;
		dp = dp ->next;
	}
	return NULL;

}/*** end getDP() ***/

/******************************************************************
 CACHE a DIR pointer for a given fd and filename
 ******************************************************************/
static
void
insertDP(unsigned long fd, char *filename, DIR *dirp)
{
	struct cpCache *dp = dpCacheHead;

	while (dp != NULL) {
		if((dp->fd == fd) && (!strcmp(dp->filename,filename))) {
			dp->dirp = dirp;
			if(verbose || syscallVerbose)
				fprintf(stderr, "insertDP changing dirp for fd = 0x%08lx, filename = '%s'\n",fd,filename);
			
			return;
		}
		dp = dp->next;
	}

	if(verbose || syscallVerbose)
		fprintf(stderr, "insertDP called to insert for a non-exiting fd, filename\n");

	if ((dp = malloc(sizeof(struct cpCache))) == NULL)
		fatal("malloc failed");

	dp->fd 	= fd;
	dp->dirp 	= dirp;
	dp->filename 	= filename;
	dp->next 	= dpCacheHead;
	dpCacheHead 	= dp;

	dp->flags	= 0xffffffff;
	dp->mode	= 0xffffffff;

}/*** end insertDP() ***/



/******************************************************************
 SET ERROR NUMBER()
 ******************************************************************/
static
void
seterrnum(int err)
{
	/*
	 * we need to stuff err into the simulator memory, translating
	 * big to little endian in the process. The location to stuff
	 * to is given by (unsigned long)errNumAddr
	 */
	unsigned int   i;
	unsigned char  *c;

	c = (unsigned char *)&err;
	c+= sizeof(int);

	for (i = 0; i< sizeof(int); i++,c--)
		MemoryAccess(Write,errNumAddr+i, c, 1);

}/*** end seterrnum() ***/



/******************************************************************
 DO BIG->LITTLE ENDIAN swap (or vice versa)
 ******************************************************************/
#define SWAPAP(TYPE,POINTER)  swapBytes(sizeof(TYPE ), (unsigned char *)POINTER);
#define SWAP(TYPE,STRUCT,MEMBER)  swapBytes(sizeof(((struct TYPE *)STRUCT)->MEMBER), (unsigned char *)&((struct TYPE *)STRUCT)->MEMBER);

static
void
swapBytes(int num_bytes, unsigned char *where)
{
	unsigned char temp;

	switch(num_bytes) {
	    case 1: return;
	    case 2:
		    temp = *where;
		    *where = *(where+1);
		    *(where+1) = temp;
		    return;
		
	    case 4: temp = *where;
		    *where = *(where+3);
		    *(where+3) = temp;
		    temp = *(where+1);
		    *(where+1) = *(where+2);
		    *(where+2) = temp;
		    return;

	    default: fatal("bad size in swapBytes()");
	}

}/*** end swapBytes() ***/

/******************************************************************
 TRANSLATE LINUX STAT BUF TO SUN
 ******************************************************************/
/* LINUX from <sys/stat.h> and <linux/types.h> */
typedef long LINUX_time_t;
struct linux_stat {
        unsigned short  st_dev;
        unsigned short  __pad1;
        unsigned long   st_ino;
        unsigned short  st_mode;
        unsigned short  st_nlink;
        unsigned short  st_uid;
        unsigned short  st_gid;
        unsigned short  st_rdev;
        unsigned short  __pad2;
        long            st_size;
        unsigned long   st_blksize;
        unsigned long   st_blocks;
        LINUX_time_t    st_atime;
        unsigned long   __unused1;
        LINUX_time_t    st_mtime;
        unsigned long   __unused2;
        LINUX_time_t    st_ctime;
        unsigned long   __unused3;
        unsigned long   __unused4;
        unsigned long   __unused5;
};
/*
SUN:
  from <sys/stat.h>, <sys/stdtypes.h> and <sys/types.h>
    typedef long            time_t;
    typedef short   dev_t;
    typedef unsigned long   ino_t;
    typedef unsigned short  uid_t;
    typedef unsigned short  gid_t;
    typedef long    off_t;
    struct  stat {
          dev_t   st_dev;
          ino_t   st_ino;
          mode_t  st_mode;
          short   st_nlink;
          uid_t   st_uid;
          gid_t   st_gid;
          dev_t   st_rdev;
          off_t   st_size;
          time_t  st_atime;
          int     st_spare1;
          time_t  st_mtime;
          int     st_spare2;
          time_t  st_ctime;
          int     st_spare3;
          long    st_blksize;
          long    st_blocks;
          long    st_spare4[2];
    };
*/

static
void
translateStatL2S(maft_t maf, unsigned long LINUX_sb, unsigned char *s_s)
{
	unsigned int  i;
	unsigned char l_s[sizeof(struct linux_stat)];

	if (LINUX_sb == 0) {
		s_s = NULL;
		return;
	}

	assert(sizeof(struct linux_stat) == sizeof(struct stat));

	/* copy the linux struct over from simulator memory */
	for (i = 0; i < sizeof(struct linux_stat); i++)
		(*maf)(Read, LINUX_sb+i, &l_s[i], 1);

	/* first, copy the linux struct to a sun struct w/o byte translation */
	assert(s_s != NULL);
	((struct stat *)s_s)->st_dev       = ((struct linux_stat *)l_s)->st_dev;
	((struct stat *)s_s)->st_ino       = ((struct linux_stat *)l_s)->st_ino;
	((struct stat *)s_s)->st_mode      = ((struct linux_stat *)l_s)->st_mode;
	((struct stat *)s_s)->st_nlink     = ((struct linux_stat *)l_s)->st_nlink;
	((struct stat *)s_s)->st_uid       = ((struct linux_stat *)l_s)->st_uid;
	((struct stat *)s_s)->st_gid       = ((struct linux_stat *)l_s)->st_gid;
	((struct stat *)s_s)->st_rdev      = ((struct linux_stat *)l_s)->st_rdev;
	((struct stat *)s_s)->st_size      = ((struct linux_stat *)l_s)->st_size;
	((struct stat *)s_s)->st_blksize   = ((struct linux_stat *)l_s)->st_blksize;
	((struct stat *)s_s)->st_blocks    = ((struct linux_stat *)l_s)->st_blocks;
	((struct stat *)s_s)->st_atime     = ((struct linux_stat *)l_s)->st_atime;
	((struct stat *)s_s)->st_mtime     = ((struct linux_stat *)l_s)->st_mtime;
	((struct stat *)s_s)->st_ctime     = ((struct linux_stat *)l_s)->st_ctime;
	((struct stat *)s_s)->st_spare1    = 0;
	((struct stat *)s_s)->st_spare2    = 0;
	((struct stat *)s_s)->st_spare3    = 0;
	((struct stat *)s_s)->st_spare4[0] = 0;
	((struct stat *)s_s)->st_spare4[1] = 0;

	/* now do appropriate byte swapping to handle big->little endian conversion */
	SWAP(stat,s_s,st_dev);
	SWAP(stat,s_s,st_ino);
	SWAP(stat,s_s,st_mode);
	SWAP(stat,s_s,st_nlink);
	SWAP(stat,s_s,st_uid);
	SWAP(stat,s_s,st_gid);
	SWAP(stat,s_s,st_rdev);
	SWAP(stat,s_s,st_size);
	SWAP(stat,s_s,st_blksize);
	SWAP(stat,s_s,st_blocks);
	SWAP(stat,s_s,st_atime);
	SWAP(stat,s_s,st_mtime);
	SWAP(stat,s_s,st_ctime);

	if(verbose || syscallVerbose) {
		fprintf(stderr, "\ttranslating stat buffer:\n");

		fprintf(stderr, "\t\tLINUX version(input):\n");
		fprintf(stderr, "\t\t\tst_dev = 0x%04hx st_ino = 0x%08lx\tst_mode = 0x%04hx\tst_nlink = 0x%04hx\n",
			((struct linux_stat *)l_s)->st_dev,     ((struct linux_stat *)l_s)->st_ino,
			((struct linux_stat *)l_s)->st_mode,    ((struct linux_stat *)l_s)->st_nlink);
		fprintf(stderr, "\t\t\tst_uid = 0x%04hx\tst_gid = 0x%04hx\tst_rdev = 0x%04hx\tst_size = 0x%08lx\n",
			((struct linux_stat *)l_s)->st_uid,     ((struct linux_stat *)l_s)->st_gid,
			((struct linux_stat *)l_s)->st_rdev,    ((struct linux_stat *)l_s)->st_size);
		fprintf(stderr, "\t\t\tst_blksize = 0x%08lx\tst_blocks = 0x%08lx\n",
			((struct linux_stat *)l_s)->st_blksize, ((struct linux_stat *)l_s)->st_blocks);
		
		fprintf(stderr, "\t\tSUN version:\n");
		fprintf(stderr, "\t\t\ttst_dev = 0x%04hx st_ino = 0x%08lx\tst_mode = 0x%04hx\tst_nlink = 0x%04hx\n",
			((struct stat *)s_s)->st_dev,     ((struct stat *)s_s)->st_ino,
			((struct stat *)s_s)->st_mode,    ((struct stat *)s_s)->st_nlink);
		fprintf(stderr, "\t\t\tst_uid = 0x%04hx\tst_gid = 0x%04hx\tst_rdev = 0x%04hx\tst_size = 0x%08lx\n",
			((struct stat *)s_s)->st_uid,     ((struct stat *)s_s)->st_gid,
			((struct stat *)s_s)->st_rdev,    ((struct stat *)s_s)->st_size);
		fprintf(stderr, "\t\t\tst_blksize = 0x%08lx\tst_blocks = 0x%08lx\n",
			((struct stat *)s_s)->st_blksize, ((struct stat *)s_s)->st_blocks);
	}
}


/******************************************************************
 TRANSLATE SUN STAT BUF TO LINUX
 ******************************************************************/
static
void
translateStatS2L(maft_t maf, unsigned char *s_s, unsigned long LINUX_sb)
{
	unsigned int  i;
	unsigned char l_s[sizeof (struct linux_stat)];

	if (s_s == NULL)
		return;

	assert(sizeof(struct linux_stat) == sizeof(struct stat));

	/* first, copy the sun struct to a linux-like struct */
	((struct linux_stat *)l_s)->st_dev     = ((struct stat *)s_s)->st_dev;
	((struct linux_stat *)l_s)->st_ino     = ((struct stat *)s_s)->st_ino;
	((struct linux_stat *)l_s)->st_mode    = ((struct stat *)s_s)->st_mode;
	((struct linux_stat *)l_s)->st_nlink   = ((struct stat *)s_s)->st_nlink;
	((struct linux_stat *)l_s)->st_uid     = ((struct stat *)s_s)->st_uid;
	((struct linux_stat *)l_s)->st_gid     = ((struct stat *)s_s)->st_gid;
	((struct linux_stat *)l_s)->st_rdev    = ((struct stat *)s_s)->st_rdev;
	((struct linux_stat *)l_s)->st_size    = ((struct stat *)s_s)->st_size;
	((struct linux_stat *)l_s)->st_blksize = ((struct stat *)s_s)->st_blksize;
	((struct linux_stat *)l_s)->st_blocks  = ((struct stat *)s_s)->st_blocks;
	((struct linux_stat *)l_s)->st_atime   = ((struct stat *)s_s)->st_atime;
	((struct linux_stat *)l_s)->st_mtime   = ((struct stat *)s_s)->st_mtime;
	((struct linux_stat *)l_s)->st_ctime   = ((struct stat *)s_s)->st_ctime;
	((struct linux_stat *)l_s)->__pad1      = 0;
	((struct linux_stat *)l_s)->__pad2      = 0;
	((struct linux_stat *)l_s)->__unused5   = 0;
	((struct linux_stat *)l_s)->__unused4   = 0;
	((struct linux_stat *)l_s)->__unused3   = 0;
	((struct linux_stat *)l_s)->__unused2   = 0;
	((struct linux_stat *)l_s)->__unused1   = 0;

	/* now do appropriate byte swapping to handle big->little endian conversion */
	SWAP(linux_stat,l_s,st_dev);
	SWAP(linux_stat,l_s,st_ino);
	SWAP(linux_stat,l_s,st_mode);
	SWAP(linux_stat,l_s,st_nlink);
	SWAP(linux_stat,l_s,st_uid);
	SWAP(linux_stat,l_s,st_gid);
	SWAP(linux_stat,l_s,st_rdev);
	SWAP(linux_stat,l_s,st_size);
	SWAP(linux_stat,l_s,st_blksize);
	SWAP(linux_stat,l_s,st_blocks);
	SWAP(linux_stat,l_s,st_atime);
	SWAP(linux_stat,l_s,st_mtime);
	SWAP(linux_stat,l_s,st_ctime);
	
	if(verbose || syscallVerbose) {
		fprintf(stderr, "\ttranslating stat buffer:\n");

		fprintf(stderr, "\t\tSUN version:(input)\n");
		fprintf(stderr, "\t\t\tst_dev = 0x%04hx st_ino = 0x%08lx\tst_mode = 0x%04hx\tst_nlink = 0x%04hx\n",
			((struct stat *)s_s)->st_dev,     ((struct stat *)s_s)->st_ino,
			((struct stat *)s_s)->st_mode,    ((struct stat *)s_s)->st_nlink);
		fprintf(stderr, "\t\t\tst_uid = 0x%04hx\tst_gid = 0x%04hx\tst_rdev = 0x%04hx\tst_size = 0x%08lx\n",
			((struct stat *)s_s)->st_uid,     ((struct stat *)s_s)->st_gid,
			((struct stat *)s_s)->st_rdev,    ((struct stat *)s_s)->st_size);
		fprintf(stderr, "\t\t\tst_blksize = 0x%08lx\tst_blocks = 0x%08lx\n",
			((struct stat *)s_s)->st_blksize, ((struct stat *)s_s)->st_blocks);
		
		fprintf(stderr, "\t\tLINUX version:\n");
		fprintf(stderr, "\t\t\ttst_dev = 0x%04hx st_ino = 0x%08lx\tst_mode = 0x%04hx\tst_nlink = 0x%04hx\n",
			((struct linux_stat *)l_s)->st_dev,     ((struct linux_stat *)l_s)->st_ino,
			((struct linux_stat *)l_s)->st_mode,    ((struct linux_stat *)l_s)->st_nlink);
		fprintf(stderr, "\t\t\tst_uid = 0x%04hx\tst_gid = 0x%04hx\tst_rdev = 0x%04hx\tst_size = 0x%08lx\n",
			((struct linux_stat *)l_s)->st_uid,     ((struct linux_stat *)l_s)->st_gid,
			((struct linux_stat *)l_s)->st_rdev,    ((struct linux_stat *)l_s)->st_size);
		fprintf(stderr, "\t\t\tst_blksize = 0x%08lx\tst_blocks = 0x%08lx\n",
			((struct linux_stat *)l_s)->st_blksize, ((struct linux_stat *)l_s)->st_blocks);
	}
	/* now copy the translated sun struct over into the simulator's linux struct memory */
	for (i = 0; i < sizeof(struct linux_stat); i++)
		(*maf)(Write, LINUX_sb+i, &l_s[i], 1);
}





/******************************************************************
 TRANSLATE a SUN dirent into a LINUX/x86 dirent
 ******************************************************************/
/* LINUX from <sys/stat.h> */
typedef long LINUX_off_t;
struct linux_dirent {
        long            d_ino;
        LINUX_off_t     d_off;
        unsigned short  d_reclen;
        char            d_name[255+1];
};
/* 
SUN:
  typedef long    off_t;
  struct  dirent {
        off_t           d_off;
        unsigned long   d_fileno;
        unsigned short  d_reclen;
        unsigned short  d_namlen;
        char            d_name[255+1];
  };
*/
static
void
translateDES2L(maft_t maf, unsigned char *s_de, unsigned long LINUX_de)
{
	unsigned char l_de[sizeof(struct linux_dirent)];
	unsigned int  i;

	((struct linux_dirent *)l_de)->d_ino 	= ((struct dirent *)s_de)->d_fileno;
	((struct linux_dirent *)l_de)->d_off 	= ((struct dirent *)s_de)->d_off;
	((struct linux_dirent *)l_de)->d_reclen= ((struct dirent *)s_de)->d_namlen;
	for(i = 0; i< 256; i++)
		((struct linux_dirent *)l_de)->d_name[i] = ((struct dirent *)s_de)->d_name[i];
	
	SWAP(linux_dirent,l_de,d_ino);
	SWAP(linux_dirent,l_de,d_off);
	SWAP(linux_dirent,l_de,d_reclen);

	/* stuff translated struct dirent into simulator memory at [sim_de] */
	for (i = 0; i < sizeof(struct linux_dirent); i++)
		(*maf)(Write, LINUX_de+i, &l_de[i], 1);
}


/******************************************************************
 TRANSLATE a SUN rlimit struct into a LINUX/x86
 ******************************************************************/
struct linux_rlimit {
        long    rlim_cur;
        long    rlim_max;
};

/*  in SUNOS:
       resource == RLIMIT_CPU,
                         RLIMIT_FSIZE,
			 RLIMIT_DATA,
			RLIMIT_STACK,
			RLIMIT_CORE,
			RLIMIT_RSS,
			or RLIMIT_NOFILE
       and
           struct rlimit {
  	        int  rlim_cur; * current (soft) limit *
       		int  rlim_max; * hard limit *
          };


       LINUX adds two resources -- RLIMIT_NPROC and RLIMIT_MEMLOCK
*/
static
void
translateRlimitL2S(maft_t maf, unsigned long LINUX_rl,unsigned char *s_rle)
{
	unsigned char l_rle[sizeof(struct linux_rlimit)];
	unsigned int  i;

	if (LINUX_rl == 0) {
		s_rle = NULL;
		return;
	}
	
	/* read the rlimit struct from simulator memory at [LINUX_rl] */
	for (i = 0; i < sizeof(struct linux_rlimit); i++)
		(*maf)(Read, LINUX_rl+i, &l_rle[i], 1);

	((struct rlimit *)s_rle)->rlim_cur = ((struct linux_rlimit *)l_rle)->rlim_cur;
	((struct rlimit *)s_rle)->rlim_max = ((struct linux_rlimit *)l_rle)->rlim_max;

	SWAP(rlimit,s_rle,rlim_cur);
	SWAP(rlimit,s_rle,rlim_max);
}

static
void
translateRlimitS2L(maft_t maf, unsigned char *s_rle, unsigned long LINUX_rl)
{
	unsigned char l_rle[sizeof(struct linux_rlimit)];
	unsigned int  i;

	if (s_rle == NULL)
		return;

	((struct linux_rlimit *)l_rle)->rlim_cur = ((struct rlimit *)s_rle)->rlim_cur;
	((struct linux_rlimit *)l_rle)->rlim_max = ((struct rlimit *)s_rle)->rlim_max;

	SWAP(linux_rlimit,l_rle,rlim_cur);
	SWAP(linux_rlimit,l_rle,rlim_max);

	/* stuff translated struct rlimit into simulator memory at [LINUX_rl] */
	for (i = 0; i < sizeof(struct linux_rlimit); i++)
		(*maf)(Write, LINUX_rl+i, &l_rle[i], 1);
}




/******************************************************************
 TRANSLATE a SUN rusage struct into LINUX/x86
 ******************************************************************/
struct LINUX_timeval {
        long    tv_sec;         /* seconds */
        long    tv_usec;        /* microseconds */
};
struct  linux_rusage {
        struct LINUX_timeval ru_utime;        /* user time used */
        struct LINUX_timeval ru_stime;        /* system time used */
        long    ru_maxrss;              /* maximum resident set size */
        long    ru_ixrss;               /* integral shared memory size */
        long    ru_idrss;               /* integral unshared data size */
        long    ru_isrss;               /* integral unshared stack size */
        long    ru_minflt;              /* page reclaims */
        long    ru_majflt;              /* page faults */
        long    ru_nswap;               /* swaps */
        long    ru_inblock;             /* block input operations */
        long    ru_oublock;             /* block output operations */
        long    ru_msgsnd;              /* messages sent */
        long    ru_msgrcv;              /* messages received */
        long    ru_nsignals;            /* signals received */
        long    ru_nvcsw;               /* voluntary context switches */
        long    ru_nivcsw;              /* involuntary quote */
};

/* 
  SUN:

    rusage struct:
      struct    rusage {
         struct timeval ru_utime; * user time used *
         struct timeval ru_stime; * system time used *
         int  ru_maxrss;          * maximum resident set size *
         int  ru_ixrss;    * currently 0 *
         int  ru_idrss;    * integral resident set size *
         int  ru_isrss;    * currently 0 *
         int  ru_minflt;   * page faults not requiring physical I/O *
         int  ru_majflt;   * page faults requiring physical I/O *
         int  ru_nswap;    * swaps *
         int  ru_inblock;  * block input operations *
         int  ru_oublock;  * block output operations *
         int  ru_msgsnd;   * messages sent *
         int  ru_msgrcv;   * messages received *
         int  ru_nsignals; * signals received *
         int  ru_nvcsw;    * voluntary context switches *
         int  ru_nivcsw;   * involuntary context switches *
      };
     struct timeval {
             long    tv_sec;         * seconds *
             long    tv_usec;        * and microseconds *
     };
*/
static
void
translateRusageL2S(maft_t maf, unsigned long LINUX_ru, unsigned char *s_rue)
{
	unsigned char l_rue[sizeof(struct linux_rusage)];
	unsigned int  i;

	if (LINUX_ru == 0) {
		s_rue = NULL;
		return;
	}

	assert(sizeof(struct linux_rusage) <= sizeof(struct rusage));

	/* read the rusage struct from simulator memory at [LINUX_ru] */
	for (i = 0; i < sizeof(struct linux_rusage); i++)
		(*maf)(Read, LINUX_ru+i, &l_rue[i], 1);

	((struct rusage *)s_rue)->ru_utime.tv_sec	= ((struct linux_rusage *)l_rue)->ru_utime.tv_sec;
	((struct rusage *)s_rue)->ru_utime.tv_usec	= ((struct linux_rusage *)l_rue)->ru_utime.tv_usec;
	((struct rusage *)s_rue)->ru_stime.tv_sec	= ((struct linux_rusage *)l_rue)->ru_stime.tv_sec;
	((struct rusage *)s_rue)->ru_stime.tv_usec	= ((struct linux_rusage *)l_rue)->ru_stime.tv_usec;
	((struct rusage *)s_rue)->ru_maxrss		= ((struct linux_rusage *)l_rue)->ru_maxrss;
	((struct rusage *)s_rue)->ru_ixrss		= ((struct linux_rusage *)l_rue)->ru_ixrss;
	((struct rusage *)s_rue)->ru_idrss		= ((struct linux_rusage *)l_rue)->ru_idrss;
	((struct rusage *)s_rue)->ru_isrss		= ((struct linux_rusage *)l_rue)->ru_isrss;
	((struct rusage *)s_rue)->ru_minflt		= ((struct linux_rusage *)l_rue)->ru_minflt;
	((struct rusage *)s_rue)->ru_majflt		= ((struct linux_rusage *)l_rue)->ru_majflt;
	((struct rusage *)s_rue)->ru_nswap		= ((struct linux_rusage *)l_rue)->ru_nswap;
	((struct rusage *)s_rue)->ru_inblock		= ((struct linux_rusage *)l_rue)->ru_inblock;
	((struct rusage *)s_rue)->ru_oublock		= ((struct linux_rusage *)l_rue)->ru_oublock;
	((struct rusage *)s_rue)->ru_msgsnd		= ((struct linux_rusage *)l_rue)->ru_msgsnd;
	((struct rusage *)s_rue)->ru_msgrcv		= ((struct linux_rusage *)l_rue)->ru_msgrcv;
	((struct rusage *)s_rue)->ru_nsignals		= ((struct linux_rusage *)l_rue)->ru_nsignals;
	((struct rusage *)s_rue)->ru_nvcsw		= ((struct linux_rusage *)l_rue)->ru_nvcsw;
	((struct rusage *)s_rue)->ru_nivcsw		= ((struct linux_rusage *)l_rue)->ru_nivcsw;

	SWAP(rusage,s_rue,ru_utime.tv_sec);
	SWAP(rusage,s_rue,ru_utime.tv_usec);
	SWAP(rusage,s_rue,ru_stime.tv_sec);
	SWAP(rusage,s_rue,ru_stime.tv_usec);
	SWAP(rusage,s_rue,ru_maxrss);
	SWAP(rusage,s_rue,ru_ixrss);
	SWAP(rusage,s_rue,ru_idrss);
	SWAP(rusage,s_rue,ru_isrss);
	SWAP(rusage,s_rue,ru_minflt);
	SWAP(rusage,s_rue,ru_majflt);
	SWAP(rusage,s_rue,ru_nswap);
	SWAP(rusage,s_rue,ru_inblock);
	SWAP(rusage,s_rue,ru_oublock);
	SWAP(rusage,s_rue,ru_msgsnd);
	SWAP(rusage,s_rue,ru_msgrcv);
	SWAP(rusage,s_rue,ru_nsignals);
	SWAP(rusage,s_rue,ru_nvcsw);
	SWAP(rusage,s_rue,ru_nivcsw);
}

static
void
translateRusageS2L(maft_t maf, unsigned char *s_rue, unsigned long LINUX_ru)
{
	unsigned char l_rue[sizeof(struct rusage)];
	unsigned int  i;

	if (s_rue == NULL)
		return;

	assert(sizeof(struct linux_rusage) <= sizeof(struct rusage));

	((struct linux_rusage *)l_rue)->ru_utime.tv_sec	= ((struct rusage *)s_rue)->ru_utime.tv_sec;
	((struct linux_rusage *)l_rue)->ru_utime.tv_usec	= ((struct rusage *)s_rue)->ru_utime.tv_usec;
	((struct linux_rusage *)l_rue)->ru_stime.tv_sec	= ((struct rusage *)s_rue)->ru_stime.tv_sec;
	((struct linux_rusage *)l_rue)->ru_stime.tv_usec	= ((struct rusage *)s_rue)->ru_stime.tv_usec;
	((struct linux_rusage *)l_rue)->ru_maxrss		= ((struct rusage *)s_rue)->ru_maxrss;
	((struct linux_rusage *)l_rue)->ru_ixrss		= ((struct rusage *)s_rue)->ru_ixrss;
	((struct linux_rusage *)l_rue)->ru_idrss		= ((struct rusage *)s_rue)->ru_idrss;
	((struct linux_rusage *)l_rue)->ru_isrss		= ((struct rusage *)s_rue)->ru_isrss;
	((struct linux_rusage *)l_rue)->ru_minflt		= ((struct rusage *)s_rue)->ru_minflt;
	((struct linux_rusage *)l_rue)->ru_majflt		= ((struct rusage *)s_rue)->ru_majflt;
	((struct linux_rusage *)l_rue)->ru_nswap		= ((struct rusage *)s_rue)->ru_nswap;
	((struct linux_rusage *)l_rue)->ru_inblock		= ((struct rusage *)s_rue)->ru_inblock;
	((struct linux_rusage *)l_rue)->ru_oublock		= ((struct rusage *)s_rue)->ru_oublock;
	((struct linux_rusage *)l_rue)->ru_msgsnd		= ((struct rusage *)s_rue)->ru_msgsnd;
	((struct linux_rusage *)l_rue)->ru_msgrcv		= ((struct rusage *)s_rue)->ru_msgrcv;
	((struct linux_rusage *)l_rue)->ru_nsignals		= ((struct rusage *)s_rue)->ru_nsignals;
	((struct linux_rusage *)l_rue)->ru_nvcsw		= ((struct rusage *)s_rue)->ru_nvcsw;
	((struct linux_rusage *)l_rue)->ru_nivcsw		= ((struct rusage *)s_rue)->ru_nivcsw;

	SWAP(linux_rusage,l_rue,ru_utime.tv_sec);
	SWAP(linux_rusage,l_rue,ru_utime.tv_usec);
	SWAP(linux_rusage,l_rue,ru_stime.tv_sec);
	SWAP(linux_rusage,l_rue,ru_stime.tv_usec);
	SWAP(linux_rusage,l_rue,ru_maxrss);
	SWAP(linux_rusage,l_rue,ru_ixrss);
	SWAP(linux_rusage,l_rue,ru_idrss);
	SWAP(linux_rusage,l_rue,ru_isrss);
	SWAP(linux_rusage,l_rue,ru_minflt);
	SWAP(linux_rusage,l_rue,ru_majflt);
	SWAP(linux_rusage,l_rue,ru_nswap);
	SWAP(linux_rusage,l_rue,ru_inblock);
	SWAP(linux_rusage,l_rue,ru_oublock);
	SWAP(linux_rusage,l_rue,ru_msgsnd);
	SWAP(linux_rusage,l_rue,ru_msgrcv);
	SWAP(linux_rusage,l_rue,ru_nsignals);
	SWAP(linux_rusage,l_rue,ru_nvcsw);
	SWAP(linux_rusage,l_rue,ru_nivcsw);

	/* stuff translated struct rusage into simulator memory at [LINUX_ru] */
	for (i = 0; i < sizeof(struct linux_rusage); i++)
		(*maf)(Write, LINUX_ru+i, &l_rue[i], 1);
}



/******************************************************************
 TRANSLATE a SUN char into a LINUX/x86 char
 ******************************************************************/
static
void
translateCharL2S(maft_t maf, unsigned long LINUX_char, unsigned char *sun_char)
{
	unsigned int i;

	for (i = 0; i < sizeof(char); i++)
		(*maf)(Read, LINUX_char+i, &sun_char[i],1);
	
	SWAPAP(char, sun_char)
}

static
void
translateCharS2L(maft_t maf, unsigned char *sun_char, unsigned long LINUX_char)
{
	unsigned int i;

	SWAPAP(char, sun_char);

	for (i = 0; i < sizeof(char); i++)
		(*maf)(Write, LINUX_char+i, &sun_char[i],1);
}

/******************************************************************
 TRANSLATE a SUN int into a LINUX/x86 int
 ******************************************************************/
static
void
translateIntL2S(maft_t maf, unsigned long LINUX_int, unsigned char *sun_int)
{
	unsigned int i;

	for (i=0; i<sizeof(int); i++)
		(*maf)(Read, LINUX_int+i, &sun_int[i],1);

	SWAPAP(int, sun_int)
}
static
void
translateIntS2L(maft_t maf, unsigned char *sun_int, unsigned long LINUX_int)
{
	unsigned int i;

	SWAPAP(int, sun_int);

	for (i=0; i<sizeof(int); i++)
		(*maf)(Write, LINUX_int+i, &sun_int[i],1);
}



/******************************************************************
 TRANSLATE a SUN termios struct into LINUX/x86
 ******************************************************************/
/*from <linux/termios.h> and <linux/types.h> */
#define LINUX_NCCS 19
typedef unsigned long LINUX_tcflag_t;
typedef unsigned char LINUX_cc_t;
struct linux_termios {
        LINUX_tcflag_t c_iflag;               /* input mode flags */
        LINUX_tcflag_t c_oflag;               /* output mode flags */
        LINUX_tcflag_t c_cflag;               /* control mode flags */
        LINUX_tcflag_t c_lflag;               /* local mode flags */
        LINUX_cc_t c_line;                    /* line discipline */
        LINUX_cc_t c_cc[LINUX_NCCS];                /* control characters */
};
/*
SUN: from <termios.h> and <sys/stdtypes.h>
#define NCCS 17
typedef  unsigned long   tcflag_t;
typedef  unsigned char   cc_t;
struct  termios {
        tcflag_t        c_iflag;        * input modes *
        tcflag_t        c_oflag;        * output modes *
        tcflag_t        c_cflag;        * control modes *
        tcflag_t        c_lflag;        * line discipline modes *
        char            c_line;         * line discipline XXX *
        cc_t            c_cc[NCCS];     * control chars *
};


LINUX:
*/
/* c_cc characters from <linux/termios.h> */
#define LINUX_VINTR     0
#define LINUX_VQUIT     1
#define LINUX_VERASE    2
#define LINUX_VKILL     3
#define LINUX_VEOF      4
#define LINUX_VTIME     5
#define LINUX_VMIN      6
#define LINUX_VSWTC     7
#define LINUX_VSTART    8
#define LINUX_VSTOP     9
#define LINUX_VSUSP     10
#define LINUX_VEOL      11
#define LINUX_VREPRINT  12
#define LINUX_VDISCARD  13
#define LINUX_VWERASE   14
#define LINUX_VLNEXT    15
#define LINUX_VEOL2     16
static
void
translateTermiosL2S(maft_t maf, unsigned long LINUX_t, unsigned char *s_t)
{
	unsigned char l_term[sizeof(struct linux_termios)];
	unsigned int  i;

	if (LINUX_t == 0) {
		s_t = NULL;
		return;
	}

	/* read the termios struct from simulator memory at [LINUX_t] */
	for (i = 0; i < sizeof(struct linux_termios); i++)
		(*maf)(Read, LINUX_t+i, &l_term[i], 1);

	((struct termios *)s_t)->c_iflag	= ((struct linux_termios *)l_term)->c_iflag;
	((struct termios *)s_t)->c_oflag	= ((struct linux_termios *)l_term)->c_oflag;
	((struct termios *)s_t)->c_cflag	= ((struct linux_termios *)l_term)->c_cflag;
	((struct termios *)s_t)->c_lflag	= ((struct linux_termios *)l_term)->c_lflag;
	((struct termios *)s_t)->c_line	= ((struct linux_termios *)l_term)->c_line;
	((struct termios *)s_t)->c_cc[VINTR]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VINTR];
	((struct termios *)s_t)->c_cc[VQUIT]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VQUIT];
	((struct termios *)s_t)->c_cc[VERASE]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VERASE];
	((struct termios *)s_t)->c_cc[VKILL]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VKILL];
	((struct termios *)s_t)->c_cc[VEOF]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VEOF];
	((struct termios *)s_t)->c_cc[VTIME]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VTIME];
	((struct termios *)s_t)->c_cc[VMIN]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VMIN];
	((struct termios *)s_t)->c_cc[VSWTCH]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VSWTC];
	((struct termios *)s_t)->c_cc[VSTART]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VSTART];
	((struct termios *)s_t)->c_cc[VSTOP]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VSTOP];
	((struct termios *)s_t)->c_cc[VSUSP]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VSUSP];
	((struct termios *)s_t)->c_cc[VEOL]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VEOL];
	((struct termios *)s_t)->c_cc[VREPRINT]=((struct linux_termios *)l_term)->c_cc[LINUX_VREPRINT];
	((struct termios *)s_t)->c_cc[VDISCARD]=((struct linux_termios *)l_term)->c_cc[LINUX_VDISCARD];
	((struct termios *)s_t)->c_cc[VWERASE]= ((struct linux_termios *)l_term)->c_cc[LINUX_VWERASE];
	((struct termios *)s_t)->c_cc[VLNEXT]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VLNEXT];
	((struct termios *)s_t)->c_cc[VEOL2]	= ((struct linux_termios *)l_term)->c_cc[LINUX_VEOL2];

	SWAP(termios,s_t,c_iflag);
	SWAP(termios,s_t,c_oflag);
	SWAP(termios,s_t,c_cflag);
	SWAP(termios,s_t,c_lflag);
}

static
void
translateTermiosS2L(maft_t maf, unsigned char * s_term, unsigned long LINUX_t)
{
	unsigned char l_t[sizeof(struct linux_termios)];
	unsigned int  i;

	if (s_term == NULL)
		return;

	((struct linux_termios *)l_t)->c_iflag	= ((struct termios *)s_term)->c_iflag;
	((struct linux_termios *)l_t)->c_oflag	= ((struct termios *)s_term)->c_oflag;
	((struct linux_termios *)l_t)->c_cflag	= ((struct termios *)s_term)->c_cflag;
	((struct linux_termios *)l_t)->c_lflag	= ((struct termios *)s_term)->c_lflag;
	((struct linux_termios *)l_t)->c_line		= ((struct termios *)s_term)->c_line;
	((struct linux_termios *)l_t)->c_cc[LINUX_VINTR]	= ((struct termios *)s_term)->c_cc[VINTR];
	((struct linux_termios *)l_t)->c_cc[LINUX_VQUIT]	= ((struct termios *)s_term)->c_cc[VQUIT];
	((struct linux_termios *)l_t)->c_cc[LINUX_VERASE]	= ((struct termios *)s_term)->c_cc[VERASE];
	((struct linux_termios *)l_t)->c_cc[LINUX_VKILL]	= ((struct termios *)s_term)->c_cc[VKILL];
	((struct linux_termios *)l_t)->c_cc[LINUX_VEOF]	= ((struct termios *)s_term)->c_cc[VEOF];
	((struct linux_termios *)l_t)->c_cc[LINUX_VTIME]	= ((struct termios *)s_term)->c_cc[VTIME];
	((struct linux_termios *)l_t)->c_cc[LINUX_VMIN]	= ((struct termios *)s_term)->c_cc[VMIN];
	((struct linux_termios *)l_t)->c_cc[LINUX_VSWTC]	= ((struct termios *)s_term)->c_cc[VSWTCH];
	((struct linux_termios *)l_t)->c_cc[LINUX_VSTART]	= ((struct termios *)s_term)->c_cc[VSTART];
	((struct linux_termios *)l_t)->c_cc[LINUX_VSTOP]	= ((struct termios *)s_term)->c_cc[VSTOP];
	((struct linux_termios *)l_t)->c_cc[LINUX_VSUSP]	= ((struct termios *)s_term)->c_cc[VSUSP];
	((struct linux_termios *)l_t)->c_cc[LINUX_VEOL]	= ((struct termios *)s_term)->c_cc[VEOL];
	((struct linux_termios *)l_t)->c_cc[LINUX_VREPRINT]	= ((struct termios *)s_term)->c_cc[VREPRINT];
	((struct linux_termios *)l_t)->c_cc[LINUX_VDISCARD]	= ((struct termios *)s_term)->c_cc[VDISCARD];
	((struct linux_termios *)l_t)->c_cc[LINUX_VWERASE]	= ((struct termios *)s_term)->c_cc[VWERASE];
	((struct linux_termios *)l_t)->c_cc[LINUX_VLNEXT]	= ((struct termios *)s_term)->c_cc[VLNEXT];
	((struct linux_termios *)l_t)->c_cc[LINUX_VEOL2]	= ((struct termios *)s_term)->c_cc[VEOL2];

	SWAP(linux_termios,l_t,c_iflag);
	SWAP(linux_termios,l_t,c_oflag);
	SWAP(linux_termios,l_t,c_cflag);
	SWAP(linux_termios,l_t,c_lflag);
	
	/* stuff translated struct termios into simulator memory at [LINUX_t] */
	for (i = 0; i < sizeof(struct linux_termios); i++)
		(*maf)(Write, LINUX_t+i, &l_t[i], 1);
}


/******************************************************************
 TRANSLATE a SUN termio struct into LINUX/x86
 ******************************************************************/
/* from <linux/termios.h> */
#define LINUX_NCC 8
struct linux_termio {
        unsigned short c_iflag;         /* input mode flags */
        unsigned short c_oflag;         /* output mode flags */
        unsigned short c_cflag;         /* control mode flags */
        unsigned short c_lflag;         /* local mode flags */
        unsigned char c_line;           /* line discipline */
        unsigned char c_cc[LINUX_NCC];        /* control characters */
};
/*
SUN: from <termio.h>
#define NCC 8
struct termio {
        unsigned short  c_iflag;        * input modes *
        unsigned short  c_oflag;        * output modes *
        unsigned short  c_cflag;        * control modes *
        unsigned short  c_lflag;        * line discipline modes *
        char    c_line;                 * line discipline *
        unsigned char   c_cc[NCC];      * control chars *
};
*/
static
void
translateTermioL2S(maft_t maf, unsigned long LINUX_t, unsigned char *s_t)
{
	unsigned char l_term[sizeof(struct linux_termio)];
	unsigned int  i;

	if (LINUX_t == 0) {
		s_t = NULL;
		return;
	}

	/* read the termio struct from simulator memory at [LINUX_t] */
	for (i = 0; i < sizeof(struct linux_termio); i++)
		(*maf)(Read, LINUX_t+i, &l_term[i], 1);

	((struct termio *)s_t)->c_iflag	= ((struct linux_termio *)l_term)->c_iflag;
	((struct termio *)s_t)->c_oflag	= ((struct linux_termio *)l_term)->c_oflag;
	((struct termio *)s_t)->c_cflag	= ((struct linux_termio *)l_term)->c_cflag;
	((struct termio *)s_t)->c_lflag	= ((struct linux_termio *)l_term)->c_lflag;
	((struct termio *)s_t)->c_line	= ((struct linux_termio *)l_term)->c_line;
	assert(LINUX_NCC == NCC);
	for (i = 0; i < LINUX_NCC; i++)
		((struct termio *)s_t)->c_cc[i]	= ((struct linux_termio *)l_term)->c_cc[i];

	SWAP(termio,s_t,c_iflag);
	SWAP(termio,s_t,c_oflag);
	SWAP(termio,s_t,c_cflag);
	SWAP(termio,s_t,c_lflag);
}

static
void
translateTermioS2L(maft_t maf, unsigned char * s_term, unsigned long LINUX_t)
{
	unsigned char l_t[sizeof(struct linux_termio)];
	unsigned int  i;

	if (s_term == NULL)
		return;

	((struct linux_termio *)l_t)->c_iflag	= ((struct termio *)s_term)->c_iflag;
	((struct linux_termio *)l_t)->c_oflag	= ((struct termio *)s_term)->c_oflag;
	((struct linux_termio *)l_t)->c_cflag	= ((struct termio *)s_term)->c_cflag;
	((struct linux_termio *)l_t)->c_lflag	= ((struct termio *)s_term)->c_lflag;
	((struct linux_termio *)l_t)->c_line	= ((struct termio *)s_term)->c_line;
	assert(LINUX_NCC == NCC);
	for (i = 0; i < LINUX_NCC; i++)
		((struct linux_termio *)l_t)->c_cc[i]	= ((struct termio *)s_term)->c_cc[i];

	SWAP(linux_termio,l_t,c_iflag);
	SWAP(linux_termio,l_t,c_oflag);
	SWAP(linux_termio,l_t,c_cflag);
	SWAP(linux_termio,l_t,c_lflag);

	/* stuff translated struct termio into simulator memory at [LINUX_t] */
	for (i = 0; i < sizeof(struct linux_termio); i++)
		(*maf)(Write, LINUX_t+i, &l_t[i], 1);
}


/******************************************************************
 TRANSLATE a SUN winsize struct into LINUX/x86
 ******************************************************************/
/*from <linux/termios.h> */
struct linux_winsize {
        unsigned short ws_row;
        unsigned short ws_col;
        unsigned short ws_xpixel;
        unsigned short ws_ypixel;
};
/*
SUN:
from <>

*/
static
void
translateWinSizeL2S(maft_t maf, unsigned long LINUX_ws, unsigned char *s_ws)
{
	unsigned char l_win[sizeof(struct linux_winsize)];
	unsigned int  i;

	if (LINUX_ws == 0) {
		s_ws = NULL;
		return;
	}

	/* read the termio struct from simulator memory at [LINUX_ws] */
	for (i = 0; i < sizeof(struct linux_winsize); i++)
		(*maf)(Read, LINUX_ws+i, &l_win[i], 1);
	
	((struct winsize *)s_ws)->ws_row	= ((struct linux_winsize *)l_win)->ws_row;
	((struct winsize *)s_ws)->ws_col	= ((struct linux_winsize *)l_win)->ws_col;
	((struct winsize *)s_ws)->ws_xpixel	= ((struct linux_winsize *)l_win)->ws_xpixel;
	((struct winsize *)s_ws)->ws_ypixel	= ((struct linux_winsize *)l_win)->ws_ypixel;
	
	SWAP(winsize,s_ws,ws_row);
	SWAP(winsize,s_ws,ws_col);
	SWAP(winsize,s_ws,ws_xpixel);
	SWAP(winsize,s_ws,ws_ypixel);
}

static
void
translateWinSizeS2L(maft_t maf, unsigned char *s_ws, unsigned long LINUX_ws)
{
	unsigned char l_win[sizeof(struct linux_winsize)];
	unsigned int  i;

	if (s_ws == NULL)
		return;

	((struct linux_winsize *)l_win)->ws_row	= ((struct winsize *)s_ws)->ws_row;
	((struct linux_winsize *)l_win)->ws_col	= ((struct winsize *)s_ws)->ws_col;
	((struct linux_winsize *)l_win)->ws_xpixel	= ((struct winsize *)s_ws)->ws_xpixel;
	((struct linux_winsize *)l_win)->ws_ypixel	= ((struct winsize *)s_ws)->ws_ypixel;

	SWAP(linux_winsize,l_win,ws_row);
	SWAP(linux_winsize,l_win,ws_col);
	SWAP(linux_winsize,l_win,ws_xpixel);
	SWAP(linux_winsize,l_win,ws_ypixel);

	/* stuff translated struct winsize into simulator memory at [LINUX_ws] */
	for (i = 0; i < sizeof(struct linux_winsize); i++)
		(*maf)(Write, LINUX_ws+i, &l_win[i], 1);
}

/******************************************************************
 TRANSLATE a LINUX/x86 open flag into SUN
 ******************************************************************/
/* LINUX from <linux/fcntl.h>                   SUN from <sys/fcntlcom.h> */
#define LINUX_O_RDONLY            00 /* 0x0000 -- SUN 0x0000 */
#define LINUX_O_WRONLY            01 /* 0x0001 -- SUN 0x0001 */
#define LINUX_O_RDWR              02 /* 0x0002 -- SUN 0x0002 */
#define LINUX_O_CREAT           0100 /* 0x0040 -- SUN 0x0200 */
#define LINUX_O_EXCL            0200 /* 0x0080 -- SUN 0x0800 */
#define LINUX_O_NOCTTY          0400 /* 0x0100 -- SUN 0x8000 */
#define LINUX_O_TRUNC          01000 /* 0x0200 -- SUN 0x0400 */
#define LINUX_O_APPEND         02000 /* 0x0400 -- SUN 0x0008 */
#define LINUX_O_NONBLOCK       04000 /* 0x0800 -- SUN 0x4000 */
#define LINUX_O_SYNC          010000 /* 0x1000 -- SUN 0x2000 */
#define LINUX_FASYNC          020000 /* 0x2000 -- SUN 0x0040 */
static
int
translateOpenFlags(unsigned long Lflags)
{
	int           translatedFlags;
	unsigned int  i;

	switch(Lflags) {
	    case LINUX_O_RDONLY:translatedFlags = O_RDONLY;	break;
	    case LINUX_O_WRONLY:translatedFlags = O_WRONLY;	break;
	    case LINUX_O_RDWR:	translatedFlags = O_RDWR;	break;
	    case LINUX_O_APPEND:translatedFlags = O_APPEND;	break;
	    case LINUX_O_CREAT:	translatedFlags = O_CREAT;	break;
	    case LINUX_O_TRUNC:	translatedFlags = O_TRUNC;	break;
	    case LINUX_O_EXCL:	translatedFlags = O_EXCL;	break;
	    case LINUX_O_NOCTTY:translatedFlags = O_NOCTTY;	break;
	    case LINUX_O_NONBLOCK:translatedFlags = O_NONBLOCK;	break;
	    case LINUX_O_SYNC:	translatedFlags = O_SYNC;	break;
	    default:
		    /* it is a compound flag */
		    assert(LINUX_O_RDONLY == 0);
		    translatedFlags = 0;
		    for(i = 0; i < 8*sizeof(unsigned long); i++) {
			    switch(Lflags&(1<<i)) {
			        case LINUX_O_RDONLY:	break;
			        case LINUX_O_WRONLY:	translatedFlags |= O_WRONLY;	break;
			        case LINUX_O_RDWR:	translatedFlags |= O_RDWR;	break;
			        case LINUX_O_APPEND:	translatedFlags |= O_APPEND;	break;
			        case LINUX_O_CREAT:	translatedFlags |= O_CREAT;	break;
			        case LINUX_O_TRUNC:	translatedFlags |= O_TRUNC;	break;
			        case LINUX_O_EXCL:	translatedFlags |= O_EXCL;	break;
			        case LINUX_O_NOCTTY:	translatedFlags |= O_NOCTTY;	break;
			        case LINUX_O_NONBLOCK:	translatedFlags |= O_NONBLOCK;	break;
			        case LINUX_O_SYNC:	translatedFlags |= O_SYNC;	break;
			        default:		fatal("unrecognized bit in open flag: i = %d",i);
			    }
		    }
	    }
	return translatedFlags;
}





/******************************************************************
 TRANSLATE a SUN timeval struct into LINUX/x86
	sun timeval: from <sys/time.h>
	struct timeval {
	        long    tv_sec;         * seconds *
	        long    tv_usec;        * and microseconds *
};
(defined above...)
struct LINUX_timeval {
        long    tv_sec;         * seconds *
        long    tv_usec;        * microseconds *
};
 ******************************************************************/

static
void
translateTimevalL2S(maft_t maf, unsigned long Ltimeval, unsigned char *st_v)
{
	unsigned char ltv[sizeof(struct LINUX_timeval)];
	unsigned int  i;

	if (Ltimeval == 0) {
		st_v = NULL;
		return;
	}
	
	/* read the Linux version from simulator memory */
	for (i = 0; i < sizeof(struct LINUX_timeval); i++)
		(*maf)(Read, Ltimeval+i, &ltv[i], 1);

	((struct timeval *)st_v)->tv_sec	= ((struct LINUX_timeval *)ltv)->tv_sec;
	((struct timeval *)st_v)->tv_usec	= ((struct LINUX_timeval *)ltv)->tv_usec;

	SWAP(timeval,st_v,tv_sec);
	SWAP(timeval,st_v,tv_usec);

}
static
void
translateTimevalS2L(maft_t maf, unsigned char * st_v, unsigned long Ltimeval)
{
	unsigned char ltv[sizeof(struct LINUX_timeval)];
	unsigned int  i;

	if (st_v == NULL)
		return;

	((struct LINUX_timeval *)ltv)->tv_sec		= ((struct timeval *)st_v)->tv_sec;
	((struct LINUX_timeval *)ltv)->tv_usec	= ((struct timeval *)st_v)->tv_usec;

	SWAP(LINUX_timeval,ltv,tv_sec);
	SWAP(LINUX_timeval,ltv,tv_usec);

	/* write the Linux version into simulator memory */
	for (i = 0; i < sizeof(struct LINUX_timeval); i++)
		(*maf)(Write, Ltimeval+i, &ltv[i], 1);
}





/******************************************************************
 TRANSLATE a SUN timezone struct into LINUX/x86
	sun timezone: from <sys/time.h>
	struct timezone {
       	    int     tz_minuteswest; * minutes west of Greenwich *
            int     tz_dsttime;     * type of dst correction *
};
where tz_dsttime = one of these from <sys/time.h>
#define DST_NONE        0       * not on dst *
#define DST_USA         1       * USA style dst *
#define DST_AUST        2       * Australian style dst *
#define DST_WET         3       * Western European dst *
#define DST_MET         4       * Middle European dst *
#define DST_EET         5       * Eastern European dst *
#define DST_CAN         6       * Canada *
#define DST_GB          7       * Great Britain and Eire *
#define DST_RUM         8       * Rumania *
#define DST_TUR         9       * Turkey *
#define DST_AUSTALT     10      * Australian style with shift in 1986 *
 ******************************************************************/
struct LINUX_timezone {
        int     tz_minuteswest; /* minutes west of Greenwich */
        int     tz_dsttime;     /* type of dst correction */
};
/* from <sys/time.h> */
#define LINUX_DST_NONE        0       /* not on dst */
#define LINUX_DST_USA         1       /* USA style dst */
#define LINUX_DST_AUST        2       /* Australian style dst */
#define LINUX_DST_WET         3       /* Western European dst */
#define LINUX_DST_MET         4       /* Middle European dst */
#define LINUX_DST_EET         5       /* Eastern European dst */
#define LINUX_DST_CAN         6       /* Canada */
#define LINUX_DST_GB          7       /* Great Britain and Eire */
#define LINUX_DST_RUM         8       /* Rumania */
#define LINUX_DST_TUR         9       /* Turkey */
#define LINUX_DST_AUSTALT     10      /* Australian style with shift in 1986 */
static
void
translateTimezoneL2S(maft_t maf, unsigned long Ltimezone, unsigned char *st_z)
{
	unsigned int  i;
	unsigned char ltz[sizeof(struct LINUX_timeval)];

	if (Ltimezone == 0) {
		st_z = NULL;
		return;
	}

	/* read the Linux version from simulator memory */
	for (i = 0; i < sizeof(struct LINUX_timezone); i++)
		(*maf)(Read, Ltimezone+i, &ltz[i], 1);

	SWAP(LINUX_timezone,ltz,tz_dsttime);
	switch(((struct LINUX_timezone *)ltz)->tz_dsttime) {
	    case LINUX_DST_NONE:	((struct timezone *)st_z)->tz_dsttime = DST_NONE;	break;
	    case LINUX_DST_USA:		((struct timezone *)st_z)->tz_dsttime = DST_USA;	break;
	    case LINUX_DST_AUST:	((struct timezone *)st_z)->tz_dsttime = DST_AUST;	break;
	    case LINUX_DST_WET:		((struct timezone *)st_z)->tz_dsttime = DST_WET;	break;
	    case LINUX_DST_MET:		((struct timezone *)st_z)->tz_dsttime = DST_MET;	break;
	    case LINUX_DST_EET:		((struct timezone *)st_z)->tz_dsttime = DST_EET;	break;
	    case LINUX_DST_CAN:		((struct timezone *)st_z)->tz_dsttime = DST_CAN;	break;
	    case LINUX_DST_GB:		((struct timezone *)st_z)->tz_dsttime = DST_GB;		break;
	    case LINUX_DST_RUM:		((struct timezone *)st_z)->tz_dsttime = DST_RUM;	break;
	    case LINUX_DST_TUR:		((struct timezone *)st_z)->tz_dsttime = DST_TUR;	break;
	    case LINUX_DST_AUSTALT:	((struct timezone *)st_z)->tz_dsttime = DST_AUSTALT;	break;
	    default: fatal("bad tz_dsttime value in translateTimezoneL2S()");
	}
	((struct timezone *)st_z)->tz_minuteswest	= ((struct LINUX_timezone *)ltz)->tz_minuteswest;

	SWAP(timezone,st_z,tz_minuteswest);
}
static
void
translateTimezoneS2L(maft_t maf, unsigned char *st_z, unsigned long Ltimezone)
{
	unsigned char ltz[sizeof(struct LINUX_timezone)];
	unsigned int  i;

	if (st_z == NULL)
		return;

	switch(((struct timezone *)st_z)->tz_dsttime) {
	    case DST_NONE:        ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_NONE;       break;
	    case DST_USA:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_USA;        break;
	    case DST_AUST:        ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_AUST;       break;
	    case DST_WET:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_WET;        break;
	    case DST_MET:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_MET;        break;
	    case DST_EET:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_EET;        break;
	    case DST_CAN:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_CAN;        break;
	    case DST_GB:          ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_GB;         break;
	    case DST_RUM:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_RUM;        break;
	    case DST_TUR:         ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_TUR;        break;
	    case DST_AUSTALT:     ((struct LINUX_timezone *)ltz)->tz_dsttime = LINUX_DST_AUSTALT;    break;
	    default: fatal("bad tz_dsttime value in translateTimezoneL2S()");
	}
	((struct LINUX_timezone *)ltz)->tz_minuteswest		= ((struct timezone *)st_z)->tz_minuteswest;
	
	SWAP(LINUX_timezone,ltz,tz_minuteswest);
	SWAP(LINUX_timezone,ltz,tz_dsttime);

	/* write the Linux version into simulator memory */
	for (i = 0; i < sizeof(struct LINUX_timezone); i++)
		(*maf)(Write, Ltimezone+i, &ltz[i], 1);
}



/******************************************************************
 TRANSLATE a SUN filedescriptor pair into LINUX/x86
 ******************************************************************/
static
void
translatePipeFDS2L(maft_t maf, unsigned char *sun_fd, unsigned long Lfds)
{
	unsigned char  lfds[2*sizeof(int)];
	unsigned int   i;

	if(sun_fd == NULL)
		fatal("NULL sun_fd given to translatePipeFDS2L");

	*((int*)&lfds[0]) = *((int *)&sun_fd[0]);
	*((int*)&lfds[4]) = *((int *)&sun_fd[4]);

	SWAPAP(int,&lfds[0]);
	SWAPAP(int,&lfds[4]);

	/* write the 2 file descriptors from simulator memory */
	for (i = 0; i < 2*sizeof(int); i++)
		(*maf)(Write, Lfds+i, &lfds[i], 1);
}



/******************************************************************
 TRANSLATE a SUN tms struct into LINUX/x86
  from <sys/stdtypes.h>:
	typedef     long            clock_t;

  from <sys/times>
	struct tms {
	        clock_t tms_utime;              * user time *
	        clock_t tms_stime;              * system time *
	        clock_t tms_cutime;             * user time, children *
	        clock_t tms_cstime;             * system time, children *
	};
 ******************************************************************/

/* from <linux/types.h> */
typedef long LINUX_clock_t;
/* struct returned by the times system call:
  from <linux/times.h>  */
struct LINUX_tms {
        LINUX_clock_t tms_utime;
        LINUX_clock_t tms_stime;
        LINUX_clock_t tms_cutime;
        LINUX_clock_t tms_cstime;
};
static
void
translateTMSS2L(maft_t maf, unsigned char *sun_tms, unsigned long Ltms)
{
	unsigned char ltms[sizeof(struct LINUX_tms)];
	unsigned int  i;

	if(sun_tms == NULL)
		fatal("NULL sun_tms given to translateTMSS2L");

	((struct LINUX_tms *)ltms)->tms_utime		= ((struct tms *)sun_tms)->tms_utime;
	((struct LINUX_tms *)ltms)->tms_stime		= ((struct tms *)sun_tms)->tms_stime;
	((struct LINUX_tms *)ltms)->tms_cutime	= ((struct tms *)sun_tms)->tms_cutime;
	((struct LINUX_tms *)ltms)->tms_cstime	= ((struct tms *)sun_tms)->tms_cstime;

	SWAP(LINUX_tms,ltms,tms_utime);
	SWAP(LINUX_tms,ltms,tms_stime);
	SWAP(LINUX_tms,ltms,tms_cutime);
	SWAP(LINUX_tms,ltms,tms_cstime);

	/* write the struct to simulator memory */
	for (i = 0; i < sizeof(struct LINUX_tms); i++)
		(*maf)(Write, Ltms+i, &ltms[i], 1);
}

/******************************************************************
 TRANSLATE a SUN tms struct into LINUX/x86
 defined above:
  typedef long LINUX_time_t;
 ******************************************************************/

static
void
translateTimeTS2L(maft_t maf, unsigned char *sun_t, unsigned long Lt)
{
	unsigned char  lt[sizeof(LINUX_time_t)];
	unsigned int   i;

	if(sun_t == NULL)
		fatal("NULL sun_t given to translateTimeTS2L");

	if(Lt == 0)
		return;

	*((LINUX_time_t *)lt) = *((time_t *)sun_t);
	SWAPAP(LINUX_time_t,lt);

	/* write the struct to simulator memory */
	for (i = 0; i < sizeof(LINUX_time_t); i++)
		(*maf)(Write, Lt+i, &lt[i], 1);
}

/* fcntl commands from <linux/fcntl.h> */
#define LINUX_F_DUPFD         0
#define LINUX_F_GETFD         1
#define LINUX_F_SETFD         2
#define LINUX_F_GETFL         3
#define LINUX_F_SETFL         4
#define LINUX_F_GETLK         5
#define LINUX_F_SETLK         6
#define LINUX_F_SETLKW        7
#define LINUX_F_SETOWN        8
#define LINUX_F_GETOWN        9
static
int
translateFcntlCommand(unsigned long command)
{
	int local_command;

	switch(command) {
	    case LINUX_F_DUPFD:		local_command = F_DUPFD;	 break;
	    case LINUX_F_GETFD:		local_command = F_GETFD;	 break;
	    case LINUX_F_SETFD:		local_command = F_SETFD;	 break;
	    case LINUX_F_GETFL:		local_command = F_GETFL;	 break;
	    case LINUX_F_SETFL:		local_command = F_SETFL;	 break;
	    case LINUX_F_GETLK:		local_command = F_GETLK;	 break;
	    case LINUX_F_SETLK:		local_command = F_SETLK;	 break;
	    case LINUX_F_SETLKW:	local_command = F_SETLKW;	 break;
	    case LINUX_F_SETOWN:	local_command = F_SETOWN;	 break;
	    case LINUX_F_GETOWN:	local_command = F_GETOWN;	 break;
	    default:			fatal("bad command to fcntl sys call");
	}
	return local_command;
}

/* fcntl arguments: LINUX:<linux/fcntl.h>; SUN:</usr/include/sys/fcntlcom.h> */
#define LINUX_FD_CLOEXEC      1                         /* 1 in SUN */
#define LINUX_F_RDLCK         0                         /* 1 in SUN */
/*#define LINUX_F_WRLCK       1  same as LINUX_FD_CLOEXEC; 2 in SUN */
#define LINUX_F_UNLCK         2                         /* 3 in SUN */
static
unsigned long
translateFcntlArgument(unsigned long argument, int local_command)
{
	unsigned long local_argument;

	switch(argument) {
	    case LINUX_F_RDLCK:
		local_argument = F_RDLCK;
		break;
	    case LINUX_F_UNLCK:
		local_argument = F_UNLCK;
		break;

		/*case LINUX_F_WRLCK:*/
	    case LINUX_FD_CLOEXEC:
		/*
		 * weird case, because LINUX uses a 1 for both
		 * FD_CLOEXEC and F_WRLCK arguments. It appears that
		 * CLOEXEC is only used with 3 commands.
		 */
		if(verbose || syscallVerbose)
			fprintf(stderr, "syscall: got the weird fcntl case\n");
		if ((local_command == F_DUPFD) || (local_command == F_GETFD) || (local_command == F_SETFD))
			local_argument = FD_CLOEXEC;
		else
			local_argument = F_WRLCK;
		break;
	    default:		fatal("bad argument to fcntl sys call");
	}
	return local_argument;
}

/* from <linux/resource.h> */
#define LINUX_RUSAGE_SELF     0
#define LINUX_RUSAGE_CHILDREN (-1)
#define LINUX_RUSAGE_BOTH     (-2)
static
int
translateWho(long lwho)
{
	int local_who;

	switch(lwho) {
	    case LINUX_RUSAGE_SELF:	local_who = RUSAGE_SELF;	break;
	    case LINUX_RUSAGE_CHILDREN:	local_who = RUSAGE_CHILDREN;	break;
		
	    case LINUX_RUSAGE_BOTH:
	    default: fatal("bad who in getrusage: 0x%08lx", lwho);
	}
	return local_who;
}
/* from <linux/resource.h> */
/* resources for get/set rlimit */
#define LINUX_RLIMIT_CPU        0
#define LINUX_RLIMIT_FSIZE      1
#define LINUX_RLIMIT_DATA       2
#define LINUX_RLIMIT_STACK      3
#define LINUX_RLIMIT_CORE       4
#define LINUX_RLIMIT_RSS        5
#define LINUX_RLIMIT_NPROC      6
#define LINUX_RLIMIT_NOFILE     7
#define LINUX_RLIMIT_MEMLOCK    8
static
int
translateResource(unsigned long resource)
{
	int local_resource;

	switch(resource) {
	    case LINUX_RLIMIT_CPU:	local_resource = RLIMIT_CPU;	break;
	    case LINUX_RLIMIT_FSIZE:	local_resource = RLIMIT_FSIZE;	break;
	    case LINUX_RLIMIT_DATA:	local_resource = RLIMIT_DATA;	break;
	    case LINUX_RLIMIT_STACK:	local_resource = RLIMIT_STACK;	break;
	    case LINUX_RLIMIT_CORE:	local_resource = RLIMIT_CORE;	break;
	    case LINUX_RLIMIT_RSS:	local_resource = RLIMIT_RSS;	break;
	    case LINUX_RLIMIT_NOFILE:	local_resource = RLIMIT_NOFILE;	break;
		
	    case LINUX_RLIMIT_NPROC:
	    case LINUX_RLIMIT_MEMLOCK:
	    default:fatal("bad resource: 0x%08lx in get/setrlimit()", resource);
	}
	return local_resource;
}

/* from <unistd.h> */
#define LINUX_R_OK      4
#define LINUX_W_OK      2
#define LINUX_X_OK      1
#define LINUX_F_OK      0
static
int
translateMode(unsigned long mode)
{
	int     local_mode = 0,
		mask = 0;

	mask = 0;
	if ((mode & (unsigned long)LINUX_F_OK) == (unsigned long)LINUX_F_OK){ local_mode |= F_OK; mask |= LINUX_F_OK; }
	if ((mode & (unsigned long)LINUX_R_OK) == (unsigned long)LINUX_R_OK){ local_mode |= R_OK; mask |= LINUX_R_OK; }
	if ((mode & (unsigned long)LINUX_W_OK) == (unsigned long)LINUX_W_OK){ local_mode |= W_OK; mask |= LINUX_W_OK; }
	if ((mode & (unsigned long)LINUX_X_OK) == (unsigned long)LINUX_X_OK){ local_mode |= X_OK; mask |= LINUX_X_OK; }

	assert((mask ^ mode) == 0);
	
	return local_mode;
}
